# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
from unittest.mock import patch

import pytest

from odfdo.element import Element
from odfdo.style import Style
from odfdo.style_props import StyleProps


def test_check_area_none():
    style = Style("paragraph", name="S1")
    assert style._check_area(None) == "paragraph"


def test_get_properties_invalid_area():
    style = Style("paragraph", name="S1")
    assert style.get_properties(area="invalid_area") is None


def test_get_properties_with_children():
    style = Style("paragraph", name="S1")
    props = Element.from_tag("style:paragraph-properties")
    style.append(props)
    # Use hex for color
    props.set_attribute("fo:color", "#0000FF")
    child = Element.from_tag("style:background-image")
    child.set_attribute("fo:position", "center")
    props.append(child)

    res = style.get_properties()
    assert res["fo:color"] == "#0000FF"
    assert isinstance(res["style:background-image"], dict)
    assert res["style:background-image"]["fo:position"] == "center"


def test_update_boolean_styles_none_values():
    props = {
        "style:text-line-through-style": "none",
        "style:text-underline-style": "none",
    }
    StyleProps._update_boolean_styles(props)
    assert props["strike"] is False
    assert props["underline"] is False


def test_get_list_style_properties():
    style = Style("paragraph", name="S1")
    res = style.get_list_style_properties()
    assert isinstance(res, dict)


def test_apply_valid_properties_no_allowed():
    props_element = Element.from_tag("style:unknown-properties")
    # use valid prefix fo
    properties = {"fo:color": "#0000FF", "invalid": 123}
    StyleProps._apply_valid_properties(
        props_element,
        "unknown_area",
        properties,
    )
    assert props_element.get_attribute("fo:color") == "#0000FF"
    assert props_element.get_attribute("invalid") is None


def test_apply_valid_properties_warn():
    props_element = Element.from_tag("style:text-properties")
    properties = {"svg:color": "#0000FF"}  # not allowed in text area
    with pytest.warns(UserWarning, match="not allowed"):
        StyleProps._apply_valid_properties(
            props_element,
            "text",
            properties,
        )


def test_apply_valid_properties_lo_extensions_off():
    with patch("odfdo.style_props.USE_LO_EXTENSIONS", False):
        props_element = Element.from_tag("style:text-properties")
        properties = {"loext:test": "value"}
        with pytest.warns(UserWarning):
            StyleProps._apply_valid_properties(
                props_element,
                "text",
                properties,
            )
        assert props_element.get_attribute("loext:test") is None


def test_set_properties_style_no_props():
    style1 = Style("paragraph", name="S1")
    style2 = Style("paragraph", name="S2")
    style1.set_properties(style=style2, area="text")
    assert style1.get_element("style:text-properties") is not None


def test_set_properties_style_with_props():
    style1 = Style("paragraph", name="S1")
    style2 = Style("paragraph", name="S2")
    style2.set_properties(color="blue", area="text")
    style1.set_properties(style=style2, area="text")
    assert style1.get_properties(area="text")["fo:color"] == "#0000FF"


def test_set_properties_none_value():
    style = Style("paragraph", name="S1")
    # fo:color is allowed in "text" area
    style.set_properties(color="blue", area="text")
    assert style.get_properties(area="text")["fo:color"] == "#0000FF"
    style.set_properties(color=None, area="text")
    assert "fo:color" not in style.get_properties(area="text")


def test_set_properties_no_args():
    style = Style("paragraph", name="S1")
    style.set_properties()
    assert style.get_element("style:paragraph-properties") is not None


def test_del_properties_no_element():
    style = Style("paragraph", name="S1")
    with pytest.raises(
        ValueError,
        match="The Properties element is non-existent",
    ):
        style.del_properties(properties=["fo:color"])


def test_del_properties_properties_none():
    style = Style("paragraph", name="S1")
    style.set_properties(color="blue", area="text")
    # Should not raise anything, but element must exist
    style.del_properties(properties=None, area="text")
    assert "fo:color" in style.get_properties(area="text")


def test_del_properties_variants():
    style = Style("paragraph", name="S1")
    style.set_properties(color="blue", area="text")
    style.del_properties(properties=["color"], area="text")
    assert "fo:color" not in style.get_properties(area="text")

    style.set_properties(color="blue", area="text")
    style.del_properties(properties=["nonexistent"], area="text")
    assert style.get_properties(area="text")["fo:color"] == "#0000FF"


def test_set_properties_with_properties_and_style():
    style1 = Style("paragraph", name="S1")
    style2 = Style("paragraph", name="S2")
    style2.set_properties(color="red", area="text")
    style1.set_properties(
        properties={"color": "blue"},
        style=style2,
        area="text",
    )
    assert style1.get_properties(area="text")["fo:color"] == "#0000FF"



import ast
import os
import shutil
import subprocess as sp
from pathlib import Path

DOC_SRC = Path(__file__).parent
ROOT = DOC_SRC.parent
SRC = DOC_SRC / "src"
RECIPES_MD = SRC / "recipes.md"
SITE = DOC_SRC / "site"
DEST = DOC_SRC / ".." / "docs"
RECIPES = ROOT / "recipes"
MKDOCS_YML = DOC_SRC / "mkdocs.yml"

# List of modules to document (manually curated)
REFERENCE_MODULES = [
    # Core modules
    "annotation",
    "body",
    "bookmark",
    "cell",
    "column",
    "config_elements",
    "const",
    "container",
    "content",
    "datatype",
    "document",
    "draw_page",
    "element",
    "element_strip",
    "element_typed",
    "elements_between",
    "form",
    "form_controls",
    "form_controls_mixins",
    "form_properties",
    "frame",
    "header",
    "header_rows",
    "image",
    "line_break",
    "link",
    "list",
    "manifest",
    "master_page",
    "meta",
    "meta_auto_reload",
    "meta_field",
    "meta_hyperlink_behaviour",
    "meta_template",
    "meta_user_defined",
    "named_range",
    "note",
    "office_forms",
    "page_layout",
    "paragraph",
    "presentation_notes",
    "reference",
    "row",
    "row_group",
    "ruby_base",
    "section",
    "settings",
    "shapes",
    "smil",
    "spacer",
    "span",
    "style",
    "style_base",
    "style_containers",
    "style_defaults",
    "style_props",
    "style_utils",
    "styles",
    "svg",
    "tab",
    "table",
    "table_cache",
    "text_index",
    "toc",
    "tracked_changes",
    "unit",
    "user_field",
    "user_field_declaration",
    "variable",
    "variable_declaration",
    "version",
    "xmlpart",
    # Mixin modules (excluding mixin_md - internal markdown mixins)
    "mixin_dc_creator",
    "mixin_dc_date",
    "mixin_link",
    "mixin_list",
    "mixin_named_range",
    "mixin_paragraph",
    "mixin_paragraph_formatted",
    "mixin_toc",
    # Utilities
    "utils",
]


def clean_dirs() -> None:
    shutil.rmtree(DEST, ignore_errors=True)
    shutil.rmtree(SITE, ignore_errors=True)
    # Ensure src is a directory
    if SRC.exists() and not SRC.is_dir():
        SRC.unlink()
    SRC.mkdir(parents=True, exist_ok=True)
    shutil.copy(ROOT / "odfdo.png", SRC)
    shutil.copy(ROOT / "README.md", SRC)
    shutil.copy(ROOT / "CHANGES.md", SRC)


def generate_reference_files() -> None:
    """Generate reference markdown files from module list."""
    SRC.mkdir(parents=True, exist_ok=True)
    for module in REFERENCE_MODULES:
        # Format title: "mixin_paragraph" -> "Mixin Paragraph"
        title = module.replace("_", " ").title()
        content = f"# {title}\n\n::: odfdo.{module}\n"
        (SRC / f"reference_{module}.md").write_text(content)
    print(f"Generated {len(REFERENCE_MODULES)} reference files")


def update_mkdocs_nav() -> None:
    """Update the nav section in mkdocs.yml with reference files."""
    content = MKDOCS_YML.read_text()

    # Build reference section
    ref_lines = ["    - Reference:"]
    for module in REFERENCE_MODULES:
        ref_lines.append(f"          - {module}: reference_{module}.md")

    ref_section = "\n".join(ref_lines)

    # Find and replace the Reference section in nav
    lines = content.splitlines()
    new_lines = []
    in_ref_section = False
    ref_section_done = False

    for line in lines:
        if line.strip().startswith("- Reference:") and not ref_section_done:
            new_lines.append(ref_section)
            in_ref_section = True
            ref_section_done = True
            continue

        if in_ref_section:
            # Check if we're leaving the reference section (next top-level item)
            if line.startswith("    - ") and "Reference" not in line:
                in_ref_section = False
                new_lines.append(line)
            elif line.strip() == "" or line.startswith("          - "):
                # Still in reference section, skip old entries
                continue
            else:
                in_ref_section = False
                new_lines.append(line)
        else:
            new_lines.append(line)

    MKDOCS_YML.write_text("\n".join(new_lines))
    print(f"Updated {MKDOCS_YML}")


def write_recipes_header() -> None:
    header = (
        "# Recipes\n\n"
        "Recipes source code is in the `/recipes` directory "
        "of `odfdo` sources.\n"
        "Most recipes are autonomous scripts doing actual modifications "
        "of ODF sample files, you can check the results in the "
        "`recipes/recipes_output` directory.\n\n"
    )
    RECIPES_MD.write_text(header)


def get_seq_docstring(path: Path) -> tuple[int, str]:
    sequence = 0
    docstring = ""
    tree = ast.parse(path.read_text())
    for node in tree.body:
        if isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Name) and target.id == "_DOC_SEQUENCE":
                    sequence = int(ast.literal_eval(node.value))
                    break
    if (
        isinstance(tree.body[0], ast.Expr)
        and isinstance(tree.body[0].value, ast.Constant)
        and isinstance(tree.body[0].value.value, str)
    ):
        docstring = tree.body[0].value.value
    return sequence, docstring


def recipes_sequence() -> dict[int, tuple[Path, str]]:
    recipes: dict[int, tuple[Path, str]] = {}
    for path in RECIPES.glob("*.py"):
        sequence, docstring = get_seq_docstring(path)
        if not sequence:
            sequence = 5000 + len(recipes)
        while sequence in recipes:
            sequence += 1
        recipes[sequence] = (path, docstring)
    return recipes


def add_recipe_blocks(recipes: dict[int, tuple[Path, str]]) -> None:
    with RECIPES_MD.open("a") as file:
        for key in sorted(recipes.keys()):
            path, docstring = recipes[key]
            print("recipe:", path.stem)
            file.write(f"## {path.stem.replace('_', ' ').capitalize()}\n\n")
            file.write(f"{docstring}\n\n")
            file.write(f'??? code "recipes/{path.name}"\n')
            file.write("    ```python\n")
            file.write(f'    {{% include "../../recipes/{path.name}" %}}\n')
            file.write("    ```\n\n")


def make_doc() -> None:
    env = dict(os.environ)
    env["NO_MKDOCS_2_WARNING"] = "1"
    result = sp.run(
        ["mkdocs", "build"],
        cwd=DOC_SRC,
        env=env,
        capture_output=True,
        text=True,
    )
    # Filter out mkdocs_autorefs warnings about ambiguous anchors
    for line in result.stdout.splitlines():
        if "WARNING" in line and "mkdocs_autorefs" in line:
            continue
        print(line)
    for line in result.stderr.splitlines():
        if "WARNING" in line and "mkdocs_autorefs" in line:
            continue
        print(line, file=__import__("sys").stderr)
    shutil.move(SITE, DEST)


def main() -> None:
    clean_dirs()
    generate_reference_files()
    update_mkdocs_nav()
    recipes = recipes_sequence()
    write_recipes_header()
    add_recipe_blocks(recipes)
    make_doc()


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          Luis Belmar-Letelier <luis@itaapy.com>
#          David Versmisse <david.versmisse@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>
from io import BytesIO
from unittest.mock import MagicMock, PropertyMock, patch

import pytest

from odfdo.const import (
    XML,
    ZIP,
)
from odfdo.document import (
    AUTOMATIC_PREFIX,
    Document,
)
from odfdo.element import Element
from odfdo.image import DrawFillImage
from odfdo.page_layout import StylePageLayout
from odfdo.style import Style


def test_insert_style_paragraph(samples):
    document = Document(samples("example.odt"))
    style = Element.from_tag(
        '<style:style style:name="custom" '
        'style:display-name="custom" '
        'style:family="paragraph" '
        'style:parent-style-name="Text">'
        '<style:paragraph-properties fo:margin-left="2cm"/>'
        '<style:text-properties fo:color="#808080" loext:opacity="100%" '
        'fo:font-size="16pt" fo:font-style="normal" '
        'style:text-underline-style="solid" '
        'style:text-underline-width="auto" '
        'style:text-underline-color="font-color" '
        'fo:font-weight="bold"/>'
        "</style:style>"
    )
    nb_styles = len(document.get_styles("paragraph"))
    document.insert_style(style)
    expected = nb_styles + 1
    result = len(document.get_styles("paragraph"))
    assert result == expected
    document.insert_style(style)
    result = len(document.get_styles("paragraph"))
    assert result == expected


def test_insert_style_date(samples):
    document = Document(samples("example.odt"))
    style = Element.from_tag(
        "<number:date-style "
        'style:name="nr-nl-date" '
        'number:automatic-order="true">'
        '<number:day number:style="long"/>'
        "<number:text>-</number:text>"
        '<number:month number:style="long"/>'
        "<number:text>-</number:text>"
        '<number:year number:style="long"/>'
        "</number:date-style>"
    )
    nb_styles = len(document.get_styles("date"))
    document.insert_style(style)
    expected = nb_styles + 1
    result = len(document.get_styles("date"))
    assert result == expected
    document.insert_style(style)
    result = len(document.get_styles("date"))
    assert result == expected
    result2 = document.get_style("date", "nr-nl-date")
    assert result2


def test_insert_style_currency(samples):
    document = Document(samples("example.odt"))
    style = Element.from_tag(
        "<number:currency-style "
        'style:name="nr-nl-currency-gt0" '
        'style:volatile="true">'
        "<number:currency-symbol "
        'number:language="nl" '
        'number:country="NL">€</number:currency-symbol>'
        "<number:text> </number:text>"
        '<number:number number:decimal-places="2"'
        ' number:min-decimal-places="2" '
        'number:min-integer-digits="1"/>'
        "</number:currency-style>"
    )
    nb_styles = len(document.get_styles("currency"))
    document.insert_style(style)
    expected = nb_styles + 1
    result = len(document.get_styles("currency"))
    assert result == expected
    document.insert_style(style)
    result = len(document.get_styles("currency"))
    assert result == expected
    result2 = document.get_style("currency", "nr-nl-currency-gt0")
    assert result2


def test_insert_style_currency_2(samples):
    document = Document(samples("example.odt"))
    style = Element.from_tag(
        "<number:currency-style "
        'style:name="nr-nl-currency">'
        "<number:currency-symbol "
        'number:language="nl" '
        'number:country="NL">€</number:currency-symbol>'
        "<number:text> </number:text>"
        '<number:number number:decimal-places="2" '
        'number:min-decimal-places="2" '
        'number:min-integer-digits="1"/>'
        "<number:text>-</number:text>"
        "<style:map "
        'style:condition="value()&gt;=0" '
        'style:apply-style-name="nr-nl-currency-gt0"/>'
        "</number:currency-style>"
    )
    nb_styles = len(document.get_styles("currency"))
    document.insert_style(style)
    expected = nb_styles + 1
    result = len(document.get_styles("currency"))
    assert result == expected
    document.insert_style(style)
    result = len(document.get_styles("currency"))
    assert result == expected
    result2 = document.get_style("currency", "nr-nl-currency")
    assert result2


def test_insert_style_composite_1(samples):
    document = Document(samples("example.odt"))
    style_date = Element.from_tag(
        "<number:date-style "
        'style:name="nr-nl-date" '
        'number:automatic-order="true">'
        '<number:day number:style="long"/>'
        "<number:text>-</number:text>"
        '<number:month number:style="long"/>'
        "<number:text>-</number:text>"
        '<number:year number:style="long"/>'
        "</number:date-style>"
    )
    document.insert_style(style_date)
    style_cur = Element.from_tag(
        "<number:currency-style "
        'style:name="nr-nl-currency">'
        "<number:currency-symbol "
        'number:language="nl" '
        'number:country="NL">€</number:currency-symbol>'
        "<number:text> </number:text>"
        '<number:number number:decimal-places="2" '
        'number:min-decimal-places="2" '
        'number:min-integer-digits="1"/>'
        "<number:text>-</number:text>"
        "<style:map "
        'style:condition="value()&gt;=0" '
        'style:apply-style-name="nr-nl-currency-gt0"/>'
        "</number:currency-style>"
    )
    document.insert_style(style_cur)
    style = Element.from_tag(
        "<style:style "
        'style:name="nldate" '
        'style:family="table-cell" '
        'style:parent-style-name="Default" '
        'style:data-style-name="nr-nl-date"/>'
    )
    nb_styles = len(document.get_styles("table-cell"))
    document.insert_style(style)
    expected = nb_styles + 1
    result = len(document.get_styles("table-cell"))
    assert result == expected
    document.insert_style(style)
    result = len(document.get_styles("table-cell"))
    assert result == expected


def test_insert_style_composite_2(samples):
    document = Document(samples("example.odt"))
    style_date = Element.from_tag(
        "<number:date-style "
        'style:name="nr-nl-date" '
        'number:automatic-order="true">'
        '<number:day number:style="long"/>'
        "<number:text>-</number:text>"
        '<number:month number:style="long"/>'
        "<number:text>-</number:text>"
        '<number:year number:style="long"/>'
        "</number:date-style>"
    )
    document.insert_style(style_date)
    style_cur = Element.from_tag(
        "<number:currency-style "
        'style:name="nr-nl-currency">'
        "<number:currency-symbol "
        'number:language="nl" '
        'number:country="NL">€</number:currency-symbol>'
        "<number:text> </number:text>"
        '<number:number number:decimal-places="2" '
        'number:min-decimal-places="2" '
        'number:min-integer-digits="1"/>'
        "<number:text>-</number:text>"
        "<style:map "
        'style:condition="value()&gt;=0" '
        'style:apply-style-name="nr-nl-currency-gt0"/>'
        "</number:currency-style>"
    )
    document.insert_style(style_cur)
    style = Element.from_tag(
        "<style:style "
        'style:name="nlcurrency" '
        'style:family="table-cell" '
        'style:parent-style-name="Default" '
        'style:data-style-name="nr-nl-currency"/>'
    )
    nb_styles = len(document.get_styles("table-cell"))
    document.insert_style(style)
    expected = nb_styles + 1
    result = len(document.get_styles("table-cell"))
    assert result == expected
    document.insert_style(style)
    result = len(document.get_styles("table-cell"))
    assert result == expected


def test_document_insert_style_errors():
    doc = Document("text")
    with pytest.raises(TypeError, match="Unknown Style type"):
        doc.insert_style(123)

    style = Element.from_tag('<style:style style:family="paragraph"/>')
    with pytest.raises(AttributeError):
        doc.insert_style(style)


def test_document_insert_style_font_face():
    doc = Document("text")
    style = Style("font-face", font_name="MyFont")
    doc.insert_style(style)
    doc.insert_style(style, default=True)


def test_document_insert_style_page_layout():
    doc = Document("text")
    style = StylePageLayout("MyLayout")
    doc.insert_style(style)


def test_document_insert_style_draw_fill_image():
    doc = Document("text")
    style = DrawFillImage(name="MyImg")
    doc.insert_style(style)


def test_insert_style_automatic_no_name_provided():
    doc = Document("text")
    style = Style("paragraph")
    # Coverage for else branch: self._set_automatic_name
    name = doc.insert_style(style, automatic=True)
    assert name.startswith(AUTOMATIC_PREFIX)


def test_merge_styles_from_stylename_none_branch():
    doc1 = Document("text")
    doc2 = Document("text")
    # Style without name but with family
    style = Element.from_tag("style:style")
    style.set_attribute("style:family", "paragraph")
    # Manually append to common styles of doc1
    doc1.styles.get_element("//office:styles").append(style)
    # stylename will be None
    doc2.merge_styles_from(doc1)
    # Should find it via family and None name
    assert doc2.styles.get_style("paragraph", None) is not None


def test_merge_styles_from_invalid_container():
    doc1 = Document("text")
    doc2 = Document("text")
    style = Style("paragraph", name="P1")
    # Put it in a wrong container name
    wrong_container = Element.from_tag("office:wrong")
    styles_part = Element.from_tag("office:document-styles")
    styles_part.append(wrong_container)
    wrong_container.append(style)

    with patch.object(Document, "get_styles", return_value=[style]):
        with pytest.raises(NotImplementedError):
            doc2.merge_styles_from(doc1)


def test_clone_no_container_branch():
    doc = Document()
    doc.container = None
    with pytest.raises(ValueError, match="Empty Container"):
        _ = doc.clone


def test_save_pretty_no_packaging_xml():
    doc = Document("text")
    out = BytesIO()
    # packaging is ZIP, pretty is True
    doc.save(target=out, pretty=True)
    assert out.getvalue() != b""


def test_save_not_pretty_xml_packaging():
    doc = Document("text")
    out = BytesIO()
    # packaging is XML, pretty is False
    doc.save(target=out, packaging=XML, pretty=False)
    assert out.getvalue() != b""


def test_insert_style_automatic_name_exists():
    doc = Document("text")
    style = Style("paragraph", name="P1")
    # Coverage for if name: in _insert_style_get_automatic_styles
    # and if hasattr(style, "name"):
    doc.insert_style(style, name="P2", automatic=True)
    assert style.name == "P2"


def test_merge_styles_from_no_family():
    doc1 = Document("text")
    doc2 = Document("text")
    style = Element.from_tag("style:style")
    # family is None
    doc1.styles.get_element("//office:styles").append(style)
    doc2.merge_styles_from(doc1)
    # Should skip it


def test_merge_styles_from_no_container():
    doc1 = Document("text")
    doc2 = Document("text")
    style = Style("paragraph", name="P1")
    # style has no parent
    with patch.object(Document, "get_styles", return_value=[style]):
        doc2.merge_styles_from(doc1)
        # Should skip it


def test_merge_styles_from_no_upper_container():
    doc1 = Document("text")
    doc2 = Document("text")
    style = Style("paragraph", name="P1")
    container = Element.from_tag("office:styles")
    # Put it in a wrong parent
    wrong = Element.from_tag("office:wrong")
    wrong.append(container)
    container.append(style)
    # style.parent is container, container.parent is wrong
    with patch.object(Document, "get_styles", return_value=[style]):
        with pytest.raises(NotImplementedError):
            doc2.merge_styles_from(doc1)


def test_get_formatted_text_footnotes_non_rst_branch():
    doc = Document("text")
    result = []
    context = {
        "footnotes": [("1", "body")],
        "annotations": [],
        "images": [],
        "endnotes": [],
    }
    doc._get_formatted_text_footnotes(result, context, rst_mode=False)
    assert "----\n" in "".join(result)
    assert "[1] body\n" in "".join(result)


def test_add_binary_part_no_container():
    doc = Document()
    doc.container = None
    with pytest.raises(ValueError, match="Empty Container"):
        doc._add_binary_part(MagicMock())


def test_clone_no_container():
    doc = Document()
    doc.container = None
    with pytest.raises(ValueError, match="Empty Container"):
        _ = doc.clone


def test_save_target_none_zip():
    doc = Document("text")
    doc.container.path = None
    with pytest.raises(
        ValueError, match="Saving a document without path requires a target"
    ):
        doc.save(target=None, packaging=ZIP)


def test_save_target_none_general():
    doc = Document("text")
    doc.container.path = None
    with pytest.raises(
        ValueError, match="Saving a document without path requires a target"
    ):
        doc.save(target=None, packaging="folder")


def test_save_backup_str():
    doc = Document("text")
    doc.container = MagicMock()
    # Mocking basic behavior for save() to reach backup check
    doc.container.get_part.return_value = b'<office:document-meta xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0" xmlns:meta="urn:oasis:names:tc:opendocument:xmlns:meta:1.0" office:version="1.2"><office:meta><meta:generator>ODFDO</meta:generator></office:meta></office:document-meta>'
    doc.save(target="some_path.odt", backup=True)
    doc.container.save.assert_called()


def test_insert_style_automatic_name_branch():
    doc = Document("text")
    style = Style("paragraph", name="P1")
    doc.insert_style(style, name="NEW_NAME", automatic=True)
    assert style.name == "NEW_NAME"


def test_insert_style_automatic_hasattr_name_false():
    doc = Document("text")
    style = Element.from_tag("style:style")
    style.set_attribute("style:family", "paragraph")
    doc.insert_style(style, name="ANY", automatic=True)


def test_insert_style_draw_fill_image_old():
    doc = Document("text")
    fill_image = DrawFillImage("test_image", "Pictures/test.png")
    name = doc.insert_style(fill_image)
    assert name == "test_image"


def test_insert_style_invalid_family():
    doc = Document("text")
    style = Style("paragraph", name="MyStyle")
    style.family = "invalid_family"
    with pytest.raises(ValueError, match="Invalid style"):
        doc.insert_style(style)


def test_insert_style_default_no_name_old():
    doc = Document("text")
    style = Style("paragraph")
    doc.insert_style(style, default=True)
    assert style.tag == "style:default-style"


def test_insert_style_combinations_old_old():
    doc = Document("text")
    style = Style("paragraph", name="MyStyle")
    with pytest.raises(AttributeError, match="Invalid combination of arguments"):
        doc.insert_style(style, automatic=True, default=True)


def test_unique_style_name_old_old():
    doc = Document("text")
    name = doc._unique_style_name("ta")
    assert name == "ta_0"
    doc.insert_style(Style("paragraph", name="ta_0"))
    name2 = doc._unique_style_name("ta")
    assert name2 == "ta_1"


def test_show_styles_draw_fill_image_old_old():
    doc = Document("text")
    fill_image = DrawFillImage("test_fill", "Pictures/test.png")
    doc.insert_style(fill_image)
    res = doc.show_styles()
    assert "test_fill" in res


def test_insert_style_automatic_with_name_old_old():
    doc = Document("text")
    style = Style("paragraph", name="P_AUTO")
    name = doc.insert_style(style, name="P_AUTO_NEW", automatic=True)
    assert name == "P_AUTO_NEW"
    assert style.name == "P_AUTO_NEW"


def test_insert_style_default_with_name_old_old():
    doc = Document("text")
    style = Style("paragraph", name="P_DEFAULT")
    doc.insert_style(style, name="IGNORED", default=True)
    assert style.tag == "style:default-style"
    assert style.get_attribute("style:name") is None


def test_insert_style_string_xml_old_old():
    doc = Document("text")
    style_xml = '<style:style style:name="P_XML" style:family="paragraph"/>'
    name = doc.insert_style(style_xml)
    assert name == "P_XML"


def test_insert_style_hasattr_name_false_branch():
    doc = Document("text")
    # Use a real Element but hide its 'name' attribute
    style = Element.from_tag("style:style")
    style.set_attribute("style:family", "paragraph")

    with patch.object(style.__class__, "name", new_callable=PropertyMock) as mock_name:
        mock_name.side_effect = AttributeError
        # Now hasattr(style, "name") is False, so style.name = name is skipped
        doc.insert_style(style, name="NEW_NAME", automatic=True)

    # Verify it was NOT set
    assert style.get_attribute("style:name") is None


def test_merge_styles_from_no_dest():
    doc1 = Document("text")
    doc2 = Document("text")
    style = Style("paragraph", name="P1")
    doc1.styles.get_element("//office:styles").append(style)
    # Mock doc2 parts directly
    doc2.styles.get_element = MagicMock(return_value=None)
    doc2.content.get_element = MagicMock(return_value=None)
    doc2.merge_styles_from(doc1)


def test_add_page_break_style_branches():
    doc = Document("text")
    # 1. False branch: exists but NOT perfect
    style = Style("paragraph", name="odfdopagebreak")
    # No break property
    style.set_properties(area="paragraph", text_align="center")
    doc.insert_style(style, automatic=False)
    doc.add_page_break_style()

    # 2. True branch: exists AND perfect
    doc.add_page_break_style()
    assert doc.get_style("paragraph", "odfdopagebreak") is not None


def test_show_styles_attribute_error():
    doc = Document("text")
    style = Style("paragraph", name="P1")
    with patch.object(Style, "name", new_callable=PropertyMock) as mock_name:
        mock_name.side_effect = AttributeError
        with patch.object(Document, "get_styles", return_value=[style]):
            with pytest.raises(AttributeError):
                doc.show_styles()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
from __future__ import annotations

import re
from collections import namedtuple
from collections.abc import Iterable
from datetime import datetime
from typing import cast

import pytest

import odfdo
import odfdo.element
from odfdo.body import Drawing
from odfdo.const import ODF_CONTENT
from odfdo.container import Container
from odfdo.element import (
    FIRST_CHILD,
    NEXT_SIBLING,
    PREV_SIBLING,
    Element,
    _decode_qname,
    _generate_odf_namespaces,
    _uri_to_prefix,
    _xpath_text_descendant_no_annotation,
    register_element_class,
    xpath_compile,
)
from odfdo.image import DrawImage
from odfdo.named_range import NamedRange
from odfdo.paragraph import Paragraph
from odfdo.section import Section
from odfdo.tracked_changes import TextChangedRegion
from odfdo.xmlpart import XmlPart

SPECIAL_CHARS = 'using < & " characters'

Sample = namedtuple("Sample", ["container", "content", "para", "anno", "span"])


class DummyElement(Element):
    _tag = "office:dummy1"


@pytest.fixture
def sample(samples) -> Iterable[Sample]:
    container = Container(samples("example.odt"))
    content = XmlPart(ODF_CONTENT, container)
    para = content.get_element("//text:p[1]")
    anno = content.get_element("//office:annotation[1]")
    span = ""
    yield Sample(container=container, content=content, para=para, anno=anno, span=span)


@pytest.fixture
def span_styles(samples) -> Iterable[Sample]:
    container = Container(samples("span_style.odt"))
    content = XmlPart(ODF_CONTENT, container)
    para = content.get_element("//text:p")
    anno = ""
    span = para.get_element("//text:span")
    yield Sample(container=container, content=content, para=para, anno=anno, span=span)


@pytest.fixture
def drawing_test() -> Iterable[tuple[Drawing, DrawImage, TextChangedRegion]]:
    container = Drawing()
    # Image with URL
    img_tag = '<draw:image xlink:href="http://example.com/img.png"/>'
    img = cast(DrawImage, Element.from_tag(img_tag))
    container.append(img)
    # Element with dc:creator and dc:date
    region = TextChangedRegion()
    info = Element.from_tag("<office:change-info/>")
    creator = Element.from_tag("<dc:creator>John Doe</dc:creator>")
    date_val = "2024-03-30T12:00:00"
    date_ele = Element.from_tag(f"<dc:date>{date_val}</dc:date>")
    info.append(creator)
    info.append(date_ele)
    region.append(info)
    container.append(region)
    yield (container, img, region)


def test_element_init_completude_vs_registry():
    all_imported = set(odfdo.__all__)
    registry = odfdo.element._class_registry
    for klass in registry.values():
        assert klass.__name__ in all_imported


def test_create_simple():
    data = "<p>Template Element</p>"
    element = Element.from_tag(data)
    assert element._canonicalize() == data


def test_create_namespace():
    data = "<text:p>Template Element</text:p>"
    element = Element.from_tag(data)
    assert element._canonicalize() == data


def test_create_qname():
    element = Element.from_tag("text:p")
    assert element._canonicalize() == "<text:p></text:p>"


def test_decode_qname_bad():
    with pytest.raises(ValueError):
        _decode_qname("hip:hop")


def test_uri_to_prefix_bad():
    with pytest.raises(ValueError):
        _uri_to_prefix("wrong_uri")


def test_etext_parent_1():
    element = Element.from_tag('<text:p style="aa"/>')
    lst = element.xpath("//@style")
    etext = lst[0]
    assert isinstance(etext.parent, Paragraph)


def test_etext_parent_2():
    element = Element.from_tag('<text:p style="aa"/>')
    lst = element.xpath("//@style")
    etext = lst[0]
    etext._EText__parent = None
    assert etext.parent is None


def test_bad_python_element():
    with pytest.raises(TypeError):
        Element("<text:p/>")


def test_bad_element_from_tag():
    with pytest.raises(TypeError):
        Element(tag_or_elem=[])


def test_make_etree_element_bad_1():
    element = Element.from_tag("<text:p/>")
    with pytest.raises(TypeError):
        element._make_etree_element([])


def test_make_etree_element_bad_2():
    element = Element.from_tag("<text:p/>")
    with pytest.raises(ValueError):
        element._make_etree_element(" ")


def test_generic_attrib_getter_bad_1():
    element = Element.from_tag("<text:p/>")
    fct = element._generic_attrib_getter("atr", "fam")
    assert fct(element) is None


def test_generic_attrib_getter_bad_2():
    element = Element.from_tag("<text:p/>")
    element.family = "nope"
    fct = element._generic_attrib_getter("atr", "fam")
    assert fct(element) is None


def test_generic_attrib_setter_bad_1():
    element = Element.from_tag("<text:p/>")
    fct = element._generic_attrib_setter("atr", "fam")
    assert fct(element, "value") is None


def test_generic_attrib_setter_bad_2():
    element = Element.from_tag("<text:p/>")
    element.family = "nope"
    fct = element._generic_attrib_setter("atr", "fam")
    assert fct(element, "value") is None


def test_make_before_regex_bad():
    element = Element.from_tag("<text:p/>")
    with pytest.raises(ValueError):
        element._make_before_regex(None, None)


def test_search_negative_position_bad_1():
    element = Element.from_tag("<text:p/>")
    pattern = re.compile("abc")
    with pytest.raises(ValueError):
        element._search_negative_position([], pattern)


def test_search_negative_position_bad_2():
    element = Element.from_tag("<text:p>abc</text:p>")
    pattern = re.compile("abc")
    with pytest.raises(TypeError):
        element._search_negative_position([element], pattern)


def test_search_positive_position_bad_1():
    element = Element.from_tag("<text:p/>")
    pattern = re.compile("abc")
    with pytest.raises(ValueError):
        element._search_positive_position([], pattern, 0)


def test_search_positive_position_bad_2():
    element = Element.from_tag("<text:p>abc</text:p>")
    pattern = re.compile("abc")
    with pytest.raises(TypeError):
        element._search_positive_position(["abc", element], pattern, 1)


def test__insert_find_text_bad():
    current = Element.from_tag("<text:p>abcde</text:p>")
    element = Element.from_tag("<text:p>fghij</text:p>")
    with pytest.raises(ValueError):
        element._insert_find_text(
            current._Element__element,
            element._Element__element,
            "x",
            "y",
            200,
            _xpath_text_descendant_no_annotation,
        )


def test_elements_repeated_sequence_1():
    para = Paragraph("aaa")
    x = xpath_compile(".")
    result = para.elements_repeated_sequence(x, "b")
    assert result == [(0, 1)]


def test_elements_repeated_sequence_2():
    para = Paragraph("aaa")
    x = xpath_compile("xxx")
    result = para.elements_repeated_sequence(x, "b")
    assert result == []


def test_get_element_list(sample):
    content_part = sample.content
    elements = content_part.get_elements("//text:p")
    # The annotation paragraph is counted
    assert len(elements) == 8


def test_get_tagname(sample):
    element = sample.para
    assert element.tag == "text:p"


def test_get_parent():
    element = Element.from_tag("<text:p><text:span/></text:p>")
    child = element.get_element("//text:span")
    parent = child.parent
    assert parent.tag == "text:p"


def test_get_root():
    element = Element.from_tag("<text:p><text:span/></text:p>")
    root = element.root
    assert root.parent is None


def test_clone(sample):
    element = sample.para
    copy = element.clone
    assert id(element) != id(copy)
    assert element.text == copy.text


def test_delete_child():
    element = Element.from_tag("<text:p><text:span/></text:p>")
    child = element.get_element("//text:span")
    element.delete(child)
    assert element._canonicalize() == "<text:p></text:p>"


def test_delete_self():
    element = Element.from_tag("<text:p><text:span/></text:p>")
    child = element.get_element("//text:span")
    child.delete()
    assert element._canonicalize() == "<text:p></text:p>"


def test_delete_self_2():
    element = Element.from_tag("<text:p><text:span/>keep</text:p>")
    child = element.get_element("//text:span")
    child.delete()
    assert element._canonicalize() == "<text:p>keep</text:p>"


def test_delete_self_3():
    element = Element.from_tag("<text:p>before<text:span/>keep</text:p>")
    child = element.get_element("//text:span")
    child.delete()
    expected = "<text:p>beforekeep</text:p>"
    assert element._canonicalize() == expected


def test_delete_self_4():
    element = Element.from_tag("<text:p><tag>x</tag>before<text:span/>keep</text:p>")
    child = element.get_element("//text:span")
    child.delete()
    expected = "<text:p><tag>x</tag>beforekeep</text:p>"
    assert element._canonicalize() == expected


def test_delete_self_5():
    element = Element.from_tag("<text:p><tag>x</tag><text:span/>keep</text:p>")
    child = element.get_element("//text:span")
    child.delete()
    expected = "<text:p><tag>x</tag>keep</text:p>"
    assert element._canonicalize() == expected


def test_delete_root():
    element = Element.from_tag("<text:p><text:span/></text:p>")
    root = element.root
    with pytest.raises(ValueError):
        root.delete()


def test_get_attributes(sample):
    element = sample.para
    attributes = element.attributes
    excepted = {"text:style-name": "Text_20_body"}
    assert attributes == excepted


def test_get_attribute(sample):
    element = sample.para
    unknown = element.get_attribute("style-name")
    assert unknown is None


def test_get_attribute_namespace(sample):
    element = sample.para
    text = element.get_attribute("text:style-name")
    assert isinstance(text, str)
    assert text == "Text_20_body"


def test_get_attribute_none(sample):
    element = sample.para
    dummy = element.get_attribute("and_now_for_sth_completely_different")
    assert dummy is None


def test_set_attribute(sample):
    element = sample.para
    element.set_attribute("test", "a value")
    assert element.get_attribute("test") == "a value"
    element.del_attribute("test")


def test_set_attribute_namespace(sample):
    element = sample.para
    element.set_attribute("text:style-name", "Note")
    assert element.get_attribute("text:style-name") == "Note"
    element.del_attribute("text:style-name")


def test_set_attribute_special(sample):
    element = sample.para
    element.set_attribute("test", SPECIAL_CHARS)
    assert element.get_attribute("test") == SPECIAL_CHARS
    element.del_attribute("test")


def test_del_attribute(sample):
    element = sample.para
    element.set_attribute("test", "test")
    element.del_attribute("test")
    assert element.get_attribute("test") is None


def test_del_attribute_raise(sample):
    element = sample.para
    with pytest.raises(KeyError):
        element.del_attribute("office:boolean-value")


def test_del_attribute_list(sample):
    element = sample.para
    element.set_attribute("test", "test")
    attr_list = ["text:style-name", "test", "notexist1", "office:boolean-value"]
    element.del_attribute_list(attr_list)
    assert element._canonicalize() == "<text:p>This is the first paragraph.</text:p>"


def test_del_attribute_namespace(sample):
    element = sample.para
    element.set_attribute("text:style-name", "Note")
    element.del_attribute("text:style-name")
    assert element.get_attribute("text:style-name") is None


def test_set_attribute_str_default_1():
    element = Element.from_tag("text:p")
    element._set_attribute_str_default("test", "value", "default_value")
    expected = '<text:p test="value"></text:p>'
    assert element._canonicalize() == expected


def test_set_attribute_str_default_2():
    element = Element.from_tag("text:p")
    element._set_attribute_str_default("test", "default_value", "default_value")
    assert element._canonicalize() == "<text:p></text:p>"


def test_get_attribute_str_default_1():
    element = Element.from_tag('<text:p test="value"/>')
    result = element._get_attribute_str_default("test", "default_value")
    assert result == "value"


def test_get_attribute_str_default_2():
    element = Element.from_tag("<text:p/>")
    result = element._get_attribute_str_default("test", "default_value")
    assert result == "default_value"


def test_set_attr_transparent_1():
    element = Element.from_tag("<text:p/>")
    element.set_attribute("dr3d:ambient-color", "blue")
    expected = '<text:p dr3d:ambient-color="#0000FF"></text:p>'
    assert element._canonicalize() == expected


def test_set_attr_transparent_2():
    element = Element.from_tag("<text:p/>")
    element.set_attribute("dr3d:ambient-color", "transparent")
    expected = '<text:p dr3d:ambient-color="transparent"></text:p>'
    assert element._canonicalize() == expected


def test_get_text(sample):
    element = sample.para
    text = element.text
    assert text == "This is the first paragraph."


def test_set_text(sample):
    element = sample.para
    old_text = element.text
    new_text = "A test"
    element.text = new_text
    assert element.text == new_text
    element.text = old_text
    assert element.text == old_text


def test_set_text_special(sample):
    element = sample.para
    old_text = element.text
    element.text = SPECIAL_CHARS
    assert element.text == SPECIAL_CHARS
    element.text = old_text


def test_get_text_content(sample):
    element = sample.anno
    text = element.text_content
    print(element._canonicalize())
    print(repr(text))
    expected = "This is an annotation.\nWith diacritical signs: éè"
    assert text == expected


def test_set_text_content(sample):
    element = sample.anno
    old_text = element.text_content
    text = "Have a break"
    element.text_content = text
    assert element.text_content == text
    element.text_content = old_text
    assert element.text_content == old_text


def test_get_parent2(sample):
    paragraph = sample.para
    parent = paragraph.parent
    assert parent.tag == "text:section"


def test_get_parent_root(sample):
    content = sample.content
    root = content.root
    parent = root.parent
    assert parent is None


def test_insert_element_first_child():
    element = Element.from_tag("<office:text><text:p/><text:p/></office:text>")
    child = Element.from_tag("<text:h/>")
    element.insert(child, FIRST_CHILD)
    expected = (
        "<office:text><text:h></text:h><text:p></text:p><text:p></text:p></office:text>"
    )
    assert element._canonicalize() == expected


def test_insert_element_last_child():
    element = Element.from_tag("<office:text><text:p/><text:p/></office:text>")
    child = Element.from_tag("<text:h/>")
    expected = (
        "<office:text><text:p></text:p><text:p></text:p><text:h></text:h></office:text>"
    )
    element.append(child)
    assert element._canonicalize() == expected


def test_insert_element_next_sibling():
    root = Element.from_tag("<office:text><text:p/><text:p/></office:text>")
    element = root.get_elements("//text:p")[0]
    sibling = Element.from_tag("<text:h/>")
    expected = (
        "<office:text><text:p></text:p><text:h></text:h><text:p></text:p></office:text>"
    )
    element.insert(sibling, NEXT_SIBLING)
    assert root._canonicalize() == expected


def test_insert_element_next_sibling_xpath_query():
    root = Element.from_tag("<office:text><text:p/><text:p/></office:text>")
    xpath_query = xpath_compile("//text:p")
    element = root.get_elements(xpath_query)[0]
    sibling = Element.from_tag("<text:h/>")
    expected = (
        "<office:text><text:p></text:p><text:h></text:h><text:p></text:p></office:text>"
    )
    element.insert(sibling, NEXT_SIBLING)
    assert root._canonicalize() == expected


def test_insert_element_prev_sibling():
    root = Element.from_tag("<office:text><text:p/><text:p/></office:text>")
    element = root.get_elements("//text:p")[0]
    sibling = Element.from_tag("<text:h/>")
    element.insert(sibling, PREV_SIBLING)
    expected = (
        "<office:text><text:h></text:h><text:p></text:p><text:p></text:p></office:text>"
    )
    assert root._canonicalize() == expected


def test_insert_element_start_1():
    root = Element.from_tag(
        "<office:text><text:p>abc</text:p><text:p>def</text:p></office:text>"
    )
    element = root.get_elements("//text:p")[0]
    item = Element.from_tag("<text:span>new</text:span>")
    element.insert(item, start=True)
    expected = (
        "<office:text>"
        "<text:p><text:span>new</text:span>abc</text:p>"
        "<text:p>def</text:p>"
        "</office:text>"
    )
    assert root._canonicalize() == expected


def test_insert_element_start_2():
    root = Element.from_tag(
        "<office:text><text:p>abc</text:p><text:p>def</text:p></office:text>"
    )
    element = root.get_elements("//text:p")[0]
    item = Element.from_tag("<text:span>new</text:span>tail")
    element.insert(item, start=True)
    expected = (
        "<office:text>"
        "<text:p><text:span>new</text:span>tailabc</text:p>"
        "<text:p>def</text:p>"
        "</office:text>"
    )
    assert root._canonicalize() == expected


def test_insert_element_bad_element():
    element = Element.from_tag("text:p")
    with pytest.raises(AttributeError):
        element.insert("<text:span/>", FIRST_CHILD)


def test_insert_element_bad_position():
    element = Element.from_tag("text:p")
    child = Element.from_tag("text:span")
    with pytest.raises(ValueError):
        element.insert(child, 999)


def test_insert_element_bad_position_2():
    element = Element.from_tag("text:p")
    child = Element.from_tag("text:span")
    with pytest.raises(ValueError):
        element._insert(child, before="a", after="b")


def test_children(sample):
    element = sample.anno
    children = element.children
    assert len(children) == 4
    child = children[0]
    assert child.tag == "dc:creator"
    child = children[-1]
    assert child.tag == "text:p"


def test_get_element_idx():
    root = Element.from_tag("<office:text><text:p/><text:p/></office:text>")
    xpath_query = xpath_compile("//text:p")
    element = root._get_element_idx(xpath_query, 1)
    assert element.tag == "text:p"


def test_get_element_idx_none():
    root = Element.from_tag("<office:text><text:p/><text:p/></office:text>")
    xpath_query = xpath_compile("//text:p")
    element = root._get_element_idx(xpath_query, 42)
    assert element is None


def test_get_attribute_integer():
    element = Element.from_tag('<text:s text:c="4" />')
    result = element.get_attribute_integer("text:c")
    assert result == 4


def test_get_attribute_integer_bad():
    element = Element.from_tag('<text:s text:c="oops" />')
    result = element.get_attribute_integer("text:c")
    assert result is None


def test_get_attribute_integer_bad_2():
    element = Element.from_tag("<text:s/>")
    result = element.get_attribute_integer("text:c")
    assert result is None


def test_set_attribute_wrong_color():
    element = Element.from_tag("<text:p/>")
    with pytest.raises(TypeError):
        element.set_attribute("dr3d:ambient-color", False)


def test_get_attribute_integer_bad_3():
    element = Element.from_tag("<text:s />")
    result = element.get_attribute_integer("oops")
    assert result is None


def test_set_text_bad():
    element = Element.from_tag("<text:p/>")
    with pytest.raises(TypeError):
        element.text = False


def test_append_element():
    element = Element.from_tag("text:p")
    element.append("f")
    element.append("oo1")
    element.append(Element.from_tag("text:line-break"))
    element.append("f")
    element.append("oo2")
    expected = "<text:p>foo1<text:line-break></text:line-break>foo2</text:p>"
    assert element._canonicalize() == expected


def test_append_wrong_item():
    element = Element.from_tag("text:p")
    with pytest.raises(TypeError):
        element._Element__append([])


def test_search_paragraph(span_styles):
    """Search text in a paragraph."""
    pos = span_styles.para.search("ère")
    expected = 4
    assert pos == expected


def test_match_span_search(span_styles):
    """Search text in a span."""
    pos = span_styles.span.search("moust")
    assert pos == 0


def test_match_inner_span_search(span_styles):
    """Search text in a span from the parent paragraph."""
    pos = span_styles.para.search("roug")
    assert pos == 29


def test_simple_regex_search(span_styles):
    """Search a simple regex."""
    pos = span_styles.para.search("che roug")
    assert pos == 25


def test_intermediate_regex_search(span_styles):
    """Search an intermediate regex."""
    pos = span_styles.para.search("moustache (blanche|rouge)")
    assert pos == 19


def test_complex_regex_search(span_styles):
    """Search a complex regex."""
    # The (?<=...) part is pointless as we don't try to get groups from
    # a MatchObject. However, it's a valid regex expression.
    pos = span_styles.para.search(r"(?<=m)(ou)\w+(che) (blan\2|r\1ge)")
    assert pos == 20


def test_compiled_regex_search(span_styles):
    """Search with a compiled pattern."""
    pattern = re.compile(r"moustache")
    pos = span_styles.para.search(pattern)
    assert pos == 19


def test_failing_match_search(span_styles):
    """Test a regex that doesn't match."""
    pos = span_styles.para.search("Le Père moustache")
    assert pos is None


def test_match_paragraph(span_styles):
    """Match text in a paragraph."""
    match = span_styles.para.match("ère")
    assert match


def test_match_span(span_styles):
    """Match text in a span."""
    match = span_styles.span.match("moust")
    assert match


def test_match_inner_span(span_styles):
    """Match text in a span from the parent paragraph."""
    match = span_styles.para.match("roug")
    assert match


def test_simple_regex(span_styles):
    """Match a simple regex."""
    match = span_styles.para.match("che roug")
    assert match


def test_intermediate_regex(span_styles):
    """Match an intermediate regex."""
    match = span_styles.para.match("moustache (blanche|rouge)")
    assert match


def test_complex_regex(span_styles):
    """Match a complex regex."""
    # The (?<=...) part is pointless as we don't try to get groups from
    # a MatchObject. However, it's a valid regex expression.
    match = span_styles.para.match(r"(?<=m)(ou)\w+(che) (blan\2|r\1ge)")
    assert match


def test_compiled_regex(span_styles):
    """Match with a compiled pattern."""
    pattern = re.compile(r"moustache")
    match = span_styles.para.match(pattern)
    assert match


def test_failing_match(span_styles):
    """Test a regex that doesn't match."""
    match = span_styles.para.match("Le Père moustache")
    assert not match


def test_count(span_styles):
    paragraph = span_styles.para
    expected = paragraph._canonicalize()
    count = paragraph.replace("ou")
    assert count == 2
    # Ensure the orignal was not altered
    assert paragraph._canonicalize() == expected


def test_replace(span_styles):
    paragraph = span_styles.para
    clone = paragraph.clone
    count = clone.replace("moustache", "barbe")
    assert count == 1
    expected = "Le Père Noël a une barbe rouge."
    assert str(clone) == expected + "\n"
    assert clone.inner_text == expected
    # Ensure the orignal was not altered
    assert clone._canonicalize() != paragraph._canonicalize()


def test_replace_bad(span_styles):
    paragraph = span_styles.para
    clone = paragraph.clone
    with pytest.raises(Exception):  # noqa: B017
        clone.replace([], "barbe")


def test_replace_simple():
    paragraph = Element.from_tag("<text:p>abc</text:p>")
    count = paragraph.replace("abc", "xyz")
    assert count == 1


def test_across_span(span_styles):
    paragraph = span_styles.para
    count = paragraph.replace("moustache rouge")
    assert count == 0


def test_missing(span_styles):
    paragraph = span_styles.para
    count = paragraph.replace("barbe")
    assert count == 0


def test_register():
    register_element_class(DummyElement)
    element = Element.from_tag("office:dummy1")
    assert isinstance(element, DummyElement)


def test_unregistered():
    element = Element.from_tag("office:dummy2")
    assert isinstance(element, Element)
    assert not isinstance(element, DummyElement)


def test_is_bound():
    paragraph = Element.from_tag("<text:p>abc</text:p>")
    assert paragraph.is_bound


def test_content_setter_empty():
    element = Element.from_tag("<office:text><text:p>abc</text:p></office:text>")
    element.text_content = None
    expected = "<office:text><text:p></text:p></office:text>"
    assert element._canonicalize() == expected


def test_is_empty_1():
    element = Element.from_tag("<text:span>abc</text:span>def")
    assert element.is_empty() is False


def test_is_empty_2():
    element = Element.from_tag("<text:span>abc</text:span>")
    assert element.is_empty() is False


def test_is_empty_3():
    element = Element.from_tag("<text:span></text:span>def")
    assert element.is_empty() is False


def test_is_empty_4():
    element = Element.from_tag("<text:span></text:span>")
    assert element.is_empty() is True


def test_xpath_str():
    element = Element.from_tag('<text:s text:c="1" />')
    result = element.xpath("//@text:c")
    assert result == ["1"]


def test_xpath_no_bool():
    element = Element.from_tag('<text:s text:c="1" />')
    result = element.xpath("boolean(//@text:c)")
    assert result == []


def test_serialize():
    element = Element.from_tag("<text:span>abc</text:span>")
    assert element._canonicalize() == "<text:span>abc</text:span>"


def test_serialize_with_ns():
    element = Element.from_tag("<text:span>abc</text:span>")
    assert element.serialize(with_ns=True) == (
        "<text:span xmlns:text="
        '"urn:oasis:names:tc:opendocument:xmlns:text:1.0">'
        "abc</text:span>"
    )


def test_set_inner_text_exists():
    element = Element.from_tag("<text:p>x <text:span>abc</text:span> y</text:p>")
    element._set_inner_text("text:span", "new")
    expected = "<text:p>x <text:span>new</text:span> y</text:p>"
    assert element._canonicalize() == expected


def test_set_inner_text_none():
    element = Element.from_tag("<text:p>x y</text:p>")
    element._set_inner_text("text:span", "new")
    expected = "<text:p>x y<text:span>new</text:span></text:p>"
    assert element._canonicalize() == expected


def test_get_inner_text_content():
    element = Element.from_tag("<text:p>x y<text:span>content</text:span></text:p>")
    result = element._get_inner_text("text:span")
    assert result == "content"


def test_get_inner_text_no_content():
    element = Element.from_tag("<text:p>x y</text:p>")
    result = element._get_inner_text("text:span")
    assert result is None


def test_sections(sample):
    content = sample.content
    result = content.body.sections
    assert len(result) == 2
    assert isinstance(result[0], Section)


@pytest.mark.parametrize(
    "tag",
    [
        "office:chart",
        "office:drawing",
        "office:presentation",
        "office:spreadsheet",
        "office:text",
    ],
)
def test_append_named_range_in_class(tag):
    nr = NamedRange("nr_name", "A1:C2", "table name é", usage="filter")
    element = Element.from_tag(tag)
    element.append_named_range(nr)
    result = element.get_named_range(name="nr_name")
    assert isinstance(result, NamedRange)


def test_element_get_attribute_int_default_1():
    element = Element.from_tag("<form:form/>")
    result = element._get_attribute_int_default("form:tab-index", 0)
    assert result == 0


def test_element_get_attribute_int_default_2():
    element = Element.from_tag("<form:form/>")
    element._set_attribute_int_default("form:tab-index", 1, 0)
    result = element._get_attribute_int_default("form:tab-index", 0)
    assert result == 1


def test_element_get_attribute_int_default_3():
    element = Element.from_tag("<form:form/>")
    element._set_attribute_int_default("form:tab-index", "not_an_int", 0)
    result = element._get_attribute_int_default("form:tab-index", 0)
    assert result == 0


def test_generate_odf_namespaces():
    original_use_lo = odfdo.element.USE_LO_EXTENSIONS
    try:
        odfdo.element.USE_LO_EXTENSIONS = False
        ns = _generate_odf_namespaces()
        assert "office" in ns
        assert "loext" not in ns

        odfdo.element.USE_LO_EXTENSIONS = True
        ns = _generate_odf_namespaces()
        assert "office" in ns
        assert "loext" in ns
    finally:
        odfdo.element.USE_LO_EXTENSIONS = original_use_lo


def test_filtered_drawing_1(drawing_test):
    drawing, img, _region = drawing_test
    # Test URL filter
    res = drawing._filtered_elements("descendant::draw:image", url=r"example\.com")
    assert len(res) == 1
    assert res[0]._canonicalize() == img._canonicalize()


def test_filtered_drawing_2(drawing_test):
    drawing, img, _region = drawing_test
    # Test URL filter
    res = drawing._filtered_elements("descendant::draw:image", url=r"example\.com")
    assert len(res) == 1
    assert res[0]._canonicalize() == img._canonicalize()


def test_filtered_drawing_3(drawing_test):
    drawing, _img, _region = drawing_test
    res = drawing._filtered_elements("descendant::draw:image", url="other")
    assert len(res) == 0


def test_filtered_drawing_4(drawing_test):
    drawing, _img, region = drawing_test
    # Test dc_creator filter
    res = drawing._filtered_elements(
        "descendant::text:changed-region", dc_creator="John"
    )
    assert len(res) == 1
    assert res[0]._canonicalize() == region._canonicalize()


def test_filtered_drawing_5(drawing_test):
    drawing, _img, _region = drawing_test
    res = drawing._filtered_elements(
        "descendant::text:changed-region", dc_creator="Jane"
    )
    assert len(res) == 0


def test_filtered_drawing_6(drawing_test):
    drawing, _img, region = drawing_test
    # Test dc_date filter
    dt = datetime(2024, 3, 30, 12, 0, 0)
    res = drawing._filtered_elements("descendant::text:changed-region", dc_date=dt)
    assert len(res) == 1
    assert res[0]._canonicalize() == region._canonicalize()


def test_filtered_drawing_7(drawing_test):
    drawing, _img, _region = drawing_test
    dt2 = datetime(2024, 3, 30, 12, 0, 1)
    res = drawing._filtered_elements("descendant::text:changed-region", dc_date=dt2)
    assert len(res) == 0



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Romain Gauthier <romain@itaapy.com>
#          Hervé Cauwelier <herve@itaapy.com>
#          David Versmisse <david.versmisse@itaapy.com>
#          Jerome Dumonteil <jerome.dumonteil@itaapy.com>
from __future__ import annotations

from collections.abc import Iterable
from unittest.mock import patch

import pytest

from odfdo.document import Document
from odfdo.element import Element
from odfdo.table import Table


@pytest.fixture
def table(samples) -> Iterable[Table]:
    # simpletable :
    #   1	1	1	2	3	3	3
    #   1	1	1	2	3	3	3
    #   1	1	1	2	3	3	3
    #   1   2	3	4	5	6	7
    document = Document(samples("simple_table.ods"))
    yield document.body.get_table(name="Example1")


@pytest.fixture
def table2(samples) -> Iterable[Table]:
    # simpletable :
    #   1	1	1	2	3	3	3
    #   1	1	1	2	3	3	3
    #   1	1	1	2	3	3	3
    #   1   2	3	4	5	6	7
    document = Document(samples("simple_table.ods"))
    table = document.body.get_table(name="Example1")
    table2 = table.clone
    table2.set_value("a1", "a")
    table2.set_value("b1", "b")
    table2.set_value("d1", "d")
    table2.set_value("b2", "")
    table2.set_value("c2", "C")
    table2.set_value("d2", "")
    yield table2


def test_span_bad1(table):
    assert table.set_span("a1:a1") is False


def test_span_sp1(table):
    table.set_span("a1:a2")
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    for coord in ("a1", "a2"):
        assert table.get_cell(coord).is_spanned() is True
    for coord in ("b1", "b2", "a3"):
        assert table.get_cell(coord).is_spanned() is False
    assert table.set_span("a1:a2") is False
    assert table.del_span("a1:a2") is True
    assert table.del_span("a1:a2") is False
    for coord in ("a1", "a2"):
        assert table.get_cell(coord).is_spanned() is False


def test_span_sp1_merge(table2):
    table = table2
    table.set_span("a1:a2", merge=True)
    # span change only display
    assert table.get_values() == [
        ["a 1", "b", 1, "d", 3, 3, 3],
        [None, "", "C", "", 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    for coord in ("a1", "a2"):
        assert table.get_cell(coord).is_spanned() is True
    for coord in ("b1", "b2", "a3"):
        assert table.get_cell(coord).is_spanned() is False
    assert table.set_span("a1:a2") is False
    assert table.del_span("a1:a2") is True
    assert table.del_span("a1:a2") is False
    for coord in ("a1", "a2"):
        assert table.get_cell(coord).is_spanned() is False


def test_span_sp2(table):
    zone = "a1:b3"
    table.set_span(zone)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [True, True, False, False, False, False, False],
        [True, True, False, False, False, False, False],
        [True, True, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_sp2_property(table):
    zone = "a1:b3"
    table.set_span(zone)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [True, True, False, False, False, False, False],
        [True, True, False, False, False, False, False],
        [True, True, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_sp2_merge(table2):
    table = table2
    zone = "a1:b3"
    table.set_span(zone, merge=True)
    # span change only display
    assert table.get_values() == [
        ["a b 1 1 1", None, 1, "d", 3, 3, 3],
        [None, None, "C", "", 3, 3, 3],
        [None, None, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [True, True, False, False, False, False, False],
        [True, True, False, False, False, False, False],
        [True, True, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_sp2_merge_property(table2):
    table = table2
    zone = "a1:b3"
    table.set_span(zone, merge=True)
    # span change only display
    assert table.get_values() == [
        ["a b 1 1 1", None, 1, "d", 3, 3, 3],
        [None, None, "C", "", 3, 3, 3],
        [None, None, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [True, True, False, False, False, False, False],
        [True, True, False, False, False, False, False],
        [True, True, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_sp3(table):
    zone = "c1:c3"
    table.set_span(zone)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, True, False, False, False, False],
        [False, False, True, False, False, False, False],
        [False, False, True, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_sp3_property(table):
    zone = "c1:c3"
    table.set_span(zone)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, True, False, False, False, False],
        [False, False, True, False, False, False, False],
        [False, False, True, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


# table 2
# [[u'a', u'b', 1, u'd', 3, 3, 3],
# [1, u'', u'C', u'', 3, 3, 3],
# [1, 1, 1, 2, 3, 3, 3],
# [1, 2, 3, 4, 5, 6, 7]])


def test_span_sp3_merge(table2):
    table = table2
    zone = "c1:c3"
    table.set_span(zone, merge=True)
    # span change only display
    assert table.get_values() == [
        ["a", "b", "1 C 1", "d", 3, 3, 3],
        [1, "", None, "", 3, 3, 3],
        [1, 1, None, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, True, False, False, False, False],
        [False, False, True, False, False, False, False],
        [False, False, True, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_sp3_merge_property(table2):
    table = table2
    zone = "c1:c3"
    table.set_span(zone, merge=True)
    # span change only display
    assert table.get_values() == [
        ["a", "b", "1 C 1", "d", 3, 3, 3],
        [1, "", None, "", 3, 3, 3],
        [1, 1, None, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, True, False, False, False, False],
        [False, False, True, False, False, False, False],
        [False, False, True, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_sp4(table):
    zone = "g1:g4"
    table.set_span(zone)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, True],
        [False, False, False, False, False, False, True],
        [False, False, False, False, False, False, True],
        [False, False, False, False, False, False, True],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_sp4_property(table):
    zone = "g1:g4"
    table.set_span(zone)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, True],
        [False, False, False, False, False, False, True],
        [False, False, False, False, False, False, True],
        [False, False, False, False, False, False, True],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_sp42(table2):
    table = table2
    zone = "g1:g4"
    table.set_span(zone, merge=True)
    # span change only display
    assert table.get_values() == [
        ["a", "b", 1, "d", 3, 3, "3 3 3 7"],
        [1, "", "C", "", 3, 3, None],
        [1, 1, 1, 2, 3, 3, None],
        [1, 2, 3, 4, 5, 6, None],
    ]
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, True],
        [False, False, False, False, False, False, True],
        [False, False, False, False, False, False, True],
        [False, False, False, False, False, False, True],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_sp42_property(table2):
    table = table2
    zone = "g1:g4"
    table.set_span(zone, merge=True)
    # span change only display
    assert table.get_values() == [
        ["a", "b", 1, "d", 3, 3, "3 3 3 7"],
        [1, "", "C", "", 3, 3, None],
        [1, 1, 1, 2, 3, 3, None],
        [1, 2, 3, 4, 5, 6, None],
    ]
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, True],
        [False, False, False, False, False, False, True],
        [False, False, False, False, False, False, True],
        [False, False, False, False, False, False, True],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_sp5(table):
    zone = "a3:c4"
    table.set_span(zone)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [True, True, True, False, False, False, False],
        [True, True, True, False, False, False, False],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_sp5_property(table):
    zone = "a3:c4"
    table.set_span(zone)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [True, True, True, False, False, False, False],
        [True, True, True, False, False, False, False],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_sp5_merge(table2):
    table = table2
    zone = "a3:c4"
    table.set_span(zone, merge=True)
    # span change only display
    assert table.get_values() == [
        ["a", "b", 1, "d", 3, 3, 3],
        [1, "", "C", "", 3, 3, 3],
        ["1 1 1 1 2 3", None, None, 2, 3, 3, 3],
        [None, None, None, 4, 5, 6, 7],
    ]
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [True, True, True, False, False, False, False],
        [True, True, True, False, False, False, False],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_sp5_merge_property(table2):
    table = table2
    zone = "a3:c4"
    table.set_span(zone, merge=True)
    # span change only display
    assert table.get_values() == [
        ["a", "b", 1, "d", 3, 3, 3],
        [1, "", "C", "", 3, 3, 3],
        ["1 1 1 1 2 3", None, None, 2, 3, 3, 3],
        [None, None, None, 4, 5, 6, 7],
    ]
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [True, True, True, False, False, False, False],
        [True, True, True, False, False, False, False],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_sp6(table):
    zone = "b3:f3"
    table.set_span(zone)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, True, True, True, True, True, False],
        [False, False, False, False, False, False, False],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_sp6_property(table):
    zone = "b3:f3"
    table.set_span(zone)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, True, True, True, True, True, False],
        [False, False, False, False, False, False, False],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_sp6_2zone(table):
    zone = "b3:f3"
    table.set_span(zone)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, True, True, True, True, True, False],
        [False, False, False, False, False, False, False],
    ]
    zone2 = "a2:a4"
    table.set_span(zone2)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [True, False, False, False, False, False, False],
        [True, True, True, True, True, True, False],
        [True, False, False, False, False, False, False],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [True, False, False, False, False, False, False],
        [True, False, False, False, False, False, False],
        [True, False, False, False, False, False, False],
    ]
    assert table.del_span(zone2) is True
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_sp6_2zone_property(table):
    zone = "b3:f3"
    table.set_span(zone)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, True, True, True, True, True, False],
        [False, False, False, False, False, False, False],
    ]
    zone2 = "a2:a4"
    table.set_span(zone2)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [True, False, False, False, False, False, False],
        [True, True, True, True, True, True, False],
        [True, False, False, False, False, False, False],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [True, False, False, False, False, False, False],
        [True, False, False, False, False, False, False],
        [True, False, False, False, False, False, False],
    ]
    assert table.del_span(zone2) is True
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
    ]


def test_span_bigger(table):
    zone = "e2:i4"
    table.set_span(zone)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 2, 3, 4, 5, 6, 7, None, None],
    ]
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, True, True, True, True, True],
        [False, False, False, False, True, True, True, True, True],
        [False, False, False, False, True, True, True, True, True],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False, False, False],
    ]


def test_span_bigger_property(table):
    zone = "e2:i4"
    table.set_span(zone)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 2, 3, 4, 5, 6, 7, None, None],
    ]
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, True, True, True, True, True],
        [False, False, False, False, True, True, True, True, True],
        [False, False, False, False, True, True, True, True, True],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False, False, False],
    ]


def test_span_bigger_merge(table):
    zone = "f4:f5"
    table.set_span(zone, merge=True)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
        [None, None, None, None, None, None, None],
    ]
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, True, False],
        [False, False, False, False, False, True],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False],
    ]


def test_span_bigger_merge_property(table):
    zone = "f4:f5"
    table.set_span(zone, merge=True)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 1, 1, 2, 3, 3, 3],
        [1, 2, 3, 4, 5, 6, 7],
        [None, None, None, None, None, None, None],
    ]
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, True, False],
        [False, False, False, False, False, True],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False],
    ]


def test_span_bigger_outside(table):
    zone = "g6:i7"
    table.set_span(zone)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 2, 3, 4, 5, 6, 7, None, None],
        [None, None, None, None, None, None, None, None, None],
        [None, None, None, None, None, None, None, None, None],
        [None, None, None, None, None, None, None, None, None],
    ]
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [],
        [False, False, False, False, False, False, True, True, True],
        [False, False, False, False, False, False, True, True, True],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [],
        [False, False, False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False, False, False],
    ]


def test_span_bigger_outside_property(table):
    zone = "g6:i7"
    table.set_span(zone)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 2, 3, 4, 5, 6, 7, None, None],
        [None, None, None, None, None, None, None, None, None],
        [None, None, None, None, None, None, None, None, None],
        [None, None, None, None, None, None, None, None, None],
    ]
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [],
        [False, False, False, False, False, False, True, True, True],
        [False, False, False, False, False, False, True, True, True],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [],
        [False, False, False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False, False, False],
    ]


def test_span_bigger_outside_merge(table):
    zone = "g6:i7"
    table.set_span(zone, merge=True)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 2, 3, 4, 5, 6, 7, None, None],
        [None, None, None, None, None, None, None, None, None],
        [None, None, None, None, None, None, None, None, None],
        [None, None, None, None, None, None, None, None, None],
    ]
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [],
        [False, False, False, False, False, False, True, True, True],
        [False, False, False, False, False, False, True, True, True],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.get_cells():
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [],
        [False, False, False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False, False, False],
    ]


def test_span_bigger_outside_merge_property(table):
    zone = "g6:i7"
    table.set_span(zone, merge=True)
    # span change only display
    assert table.get_values() == [
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 1, 1, 2, 3, 3, 3, None, None],
        [1, 2, 3, 4, 5, 6, 7, None, None],
        [None, None, None, None, None, None, None, None, None],
        [None, None, None, None, None, None, None, None, None],
        [None, None, None, None, None, None, None, None, None],
    ]
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [],
        [False, False, False, False, False, False, True, True, True],
        [False, False, False, False, False, False, True, True, True],
    ]
    assert table.del_span(zone) is True
    res = []
    for r in table.cells:
        test_row = []
        for cell in r:
            test_row.append(cell.is_spanned())
        res.append(test_row)
    assert res == [
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False],
        [],
        [False, False, False, False, False, False, False, False, False],
        [False, False, False, False, False, False, False, False, False],
    ]


def test_set_span_malformed():
    table = Table("T")
    # convert_coordinates("1") -> (None, 0)
    with pytest.raises(ValueError):
        table.set_span("1:B2")
    # "A:B2" -> (0, None, 1, 1)
    with pytest.raises(ValueError):
        table.set_span("A:B2")
    # "A1:B" -> (0, 0, 1, None)
    with pytest.raises(ValueError):
        table.set_span("A1:B")
    # "A1:2" -> (0, 0, None, 1)
    with pytest.raises(ValueError):
        table.set_span("A1:2")


def test_del_span_malformed():
    table = Table("T")
    with pytest.raises(ValueError):
        table.del_span("1")
    with pytest.raises(ValueError):
        table.del_span("A")


def test_del_span_no_span():
    table = Table("Test")
    table.set_value("A1", "v1")
    assert table.del_span("A1") is False


def test_del_span_no_rows_spanned():
    table = Table("Test")
    table.set_value("A1", "v1")
    cell = table.get_cell("A1", clone=False)
    cell.set_attribute("table:number-columns-spanned", 2)
    # table:number-rows-spanned is missing.
    assert table.del_span("A1") is False


def test_set_span_merge_single():
    table = Table("Test")
    table.set_value("A1", "v1")
    # Only A1 is non-empty.
    table.set_span("A1:B1", merge=True)
    assert table.get_value("A1") == "v1"


def test_set_span_merge_non_string():
    table = Table("Test")
    table.set_value("A1", 123)
    table.set_value("B1", 456)
    table.set_span("A1:B1", merge=True)
    assert table.get_value("A1") == "123 456"


def test_set_span_merge_branch():
    table = Table("Test")
    table.set_value("A1", "v1")
    table.set_value("B1", "v2")
    table.set_span("A1:B1", merge=True)
    assert table.get_value("A1") == "v1 v2"


def test_set_span_merge_none_mock():
    table = Table("Test")
    table.set_value("A1", "v1")
    table.set_value("B1", "v2")
    cell1 = table.get_cell("A1")
    with patch.object(Table, "get_cells", return_value=[cell1, None]):
        table.set_span("A1:B1", merge=True)


def test_set_span_branch_gaps():
    table = Table("T")
    table.set_value("A1", "v1")
    table.set_value("B1", "v2")
    # We need to mock get_cells to return [cell, None]
    c1 = table.get_cell("A1")
    with patch.object(Table, "get_cells", return_value=[c1, None]):
        table.set_span("A1:B1", merge=True)


def test_set_span_merge_val_is_not_none_branch():
    table = Table("T")
    table.set_value("A1", "v1")
    table.set_value("B1", "v2")
    table.set_span("A1:B1", merge=True)
    assert table.get_value("A1") == "v1 v2"


def test_set_span_merge_val_is_none_branch():
    table = Table("T")
    table.set_value("A1", "v1")
    table.set_value("B1", "v2")
    # We need a cell that is NOT empty aggressively
    # but get_value() is None.
    c2 = table.get_cell("B1", clone=False)
    c2.set_value(None)
    # Add a child so it's not empty aggressively
    c2.append(Element.from_tag("text:p"))
    table.set_span("A1:B1", merge=True)
    assert table.get_value("A1") == "v1"


def test_set_span_one_cell():
    table = Table("Test")
    assert table.set_span("A1") is False


def test_delete_span_single_coord():
    table = Table("Test")
    table.set_span("A1:B2")
    assert table.del_span("A1") is True
    cell = table.get_cell("A1")
    assert cell.get_attribute("table:number-columns-spanned") is None



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
from __future__ import annotations

from odfdo.element import Element
from odfdo.element_strip import strip_elements, strip_tags


def test_strip_elements():
    element = Element.from_tag("<text:p>abc<text:line-break/>xyz</text:p>")
    to_remove = element.get_element("//text:line-break")
    strip_elements(element, to_remove)
    expected = "<text:p>abcxyz</text:p>"
    assert element._canonicalize() == expected


def test_strip_elements_empty():
    element = Element.from_tag("<text:p>abc</text:p>")
    res = strip_elements(element, [])
    assert res == element


def test_strip_elements_list():
    element = Element.from_tag("<text:p>abc<text:line-break/>xyz</text:p>")
    to_remove = element.get_element("//text:line-break")
    strip_elements(element, [to_remove])
    expected = "<text:p>abcxyz</text:p>"
    assert element._canonicalize() == expected


def test_strip_elements_span():
    element = Element.from_tag("<text:p>abc <text:span>inner</text:span> xyz</text:p>")
    to_remove = element.get_element("//text:span")
    strip_elements(element, to_remove)
    expected = "<text:p>abc inner xyz</text:p>"
    assert element._canonicalize() == expected


def test_strip_tags_none():
    element = Element.from_tag("<text:p>abc</text:p>")
    res = strip_tags(element, strip=None)
    assert res == element


def test_strip_tags_protect():
    # Protection applies to the element itself (it won't be stripped)
    # and its descendants (they won't be stripped if they are in the strip list)
    element = Element.from_tag(
        "<text:p>abc <text:span>in <text:span>deep</text:span></text:span></text:p>"
    )
    # protect both spans
    res = strip_tags(element, strip=["text:span"], protect=["text:p", "text:span"])
    expected = (
        "<text:p>abc <text:span>in <text:span>deep</text:span></text:span></text:p>"
    )
    assert res._canonicalize() == expected


def test_strip_tags_default_wrap_complex():
    # Test wrapping with multiple content types
    element = Element.from_tag(
        "<text:span>abc<text:other>inner</text:other>xyz</text:span>"
    )
    res = strip_tags(element, strip=["text:span"], default="text:p")
    # Result should be ["abc", <text:other>inner</text:other>, "xyz"]
    # All parts should be preserved.
    assert res.tag == "text:p"
    expected = "<text:p>abc<text:other>inner</text:other>xyz</text:p>"
    assert res._canonicalize() == expected


def test_strip_tags_default_none():
    element = Element.from_tag("<text:span>inner</text:span>")
    # strip the top level element, no wrapping
    res = strip_tags(element, strip=["text:span"], default=None)
    assert isinstance(res, list)
    assert res == ["inner"]


def test_strip_tags_with_tail():
    element = Element.from_tag("<text:p>abc<text:span>inner</text:span>tail</text:p>")
    res = strip_tags(element, strip=["text:span"])
    assert res._canonicalize() == "<text:p>abcinnertail</text:p>"


def test_strip_tags_modified_parent():
    # Test modified branch line 158+
    # Use a valid ODF attribute
    element = Element.from_tag(
        '<text:p text:style-name="s1">abc<text:span>inner</text:span></text:p>'
    )
    # This will modify the paragraph (its child span is removed)
    res = strip_tags(element, strip=["text:span"])
    expected = '<text:p text:style-name="s1">abcinner</text:p>'
    assert res._canonicalize() == expected


def test_strip_tags_with_text_and_tail():
    # Case 1: Strip element that HAS text (hits 147-148)
    element = Element.from_tag("<text:p><text:span>text</text:span></text:p>")
    res = strip_tags(element, strip=["text:span"])
    assert res._canonicalize() == "<text:p>text</text:p>"

    # Case 2: parent has text and tail, children modified
    # <p>ParaText<span>Inner</span></p>ParaTail
    parent = Element.from_tag("<text:p>ParaText<text:span>Inner</text:span></text:p>")
    parent.tail = "ParaTail"
    # strip span
    res = strip_tags(parent, strip=["text:span"])
    # _canonicalize() does not include tail by default
    assert res._canonicalize() == "<text:p>ParaTextInner</text:p>"
    assert res.text == "ParaTextInner"
    assert res.tail == "ParaTail"


def test_strip_tags_bad_attribute(capsys):
    # Test ValueError in set_attribute
    class BadElement(Element):
        _trigger = False

        def set_attribute(self, name, value):
            if self._trigger:
                raise ValueError("Bad")
            super().set_attribute(name, value)

    # We need the parent to be BadElement and have a child to strip
    parent = BadElement(tag="text:p")
    parent.append(Element.from_tag("<text:span>inner</text:span>"))
    parent.set_attribute("text:style-name", "val")
    parent._trigger = True

    # Trigger stripping of span
    strip_tags(parent, strip=["text:span"])

    captured = capsys.readouterr()
    assert "strip_tags(): bad attribute" in captured.err



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
from __future__ import annotations

from odfdo.row import Row
from odfdo.table import Table


def test_optimize_width():
    table = Table("TestTable")
    table.set_value("A1", "content")
    table.set_value("A2", "content")
    table.set_value("B1", "content")
    table.set_value("E10", None)
    table.optimize_width()
    assert table.width == 2
    assert table.height >= 2


def test_optimize_width_many_empty_rows():
    table = Table("Test")
    table.set_value("A1", "v1")
    # append 3 empty rows
    table.append_row(Row())
    table.append_row(Row())
    table.append_row(Row())
    # Should keep one empty row
    table.optimize_width()
    assert table.height == 2


def test_optimize_width_all_empty():
    table = Table("Empty")
    table.append_row(Row())
    table.append_row(Row())
    # count == self.height (2)
    table.optimize_width()
    assert table.height == 1


def test_optimize_width_cols_internal():
    table = Table("Test", width=2, height=1)
    # width=2 creates columns.
    table.optimize_width()
    assert table.width == 2


def test_optimize_width_length_internal():
    table = Table("Test")
    table.set_value("A1", "v1")
    table.optimize_width()
    assert table.width == 1


def test_optimize_width_all_empty_try_except():
    table = Table("Empty")
    table.optimize_width()
    assert table.height == 0



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
from unittest.mock import MagicMock, patch

from odfdo.annotation import Annotation
from odfdo.element import Element
from odfdo.header import Header
from odfdo.mixin_paragraph_formatted import (
    ParaFormattedTextMixin,
    _add_object_text,
    _add_object_text_annotation,
    _add_object_text_line_break,
    _add_object_text_note,
    _add_object_text_note_end,
    _add_object_text_note_foot,
    _add_object_text_span,
    _bold_styled,
    _formatted_text,
    _italic_styled,
    _post,
    _pre,
)
from odfdo.note import Note


class DummyElement(Element, ParaFormattedTextMixin):
    pass


def test_pre_no_match():
    # Patch the regex objects themselves
    with patch("odfdo.mixin_paragraph_formatted.RE_SP_PRE") as mock_re:
        mock_re.match.return_value = None
        assert _pre("abc") == ""
    assert _pre("  abc") == "  "


def test_post_no_match():
    with patch("odfdo.mixin_paragraph_formatted.RE_SP_POST") as mock_re:
        mock_re.search.return_value = None
        assert _post("abc") == ""
    assert _post("abc  ") == "  "


def test_bold_styled():
    assert _bold_styled("  abc  ") == "  **abc**  "


def test_italic_styled():
    assert _italic_styled("  abc  ") == "  *abc*  "


def test_get_formatted_text_simple_false():
    elem = DummyElement.from_tag("text:p")
    elem.text = "hello"
    assert elem.get_formatted_text() == "hello\n\n"


def test_add_object_text_span_no_rst():
    obj = Element.from_tag("text:span")
    obj.text = "bold"
    context = {"rst_mode": False}
    result = []
    _add_object_text(obj, context, result)
    assert result == ["bold"]


def test_add_object_text_span_empty_text():
    obj = Element.from_tag("text:span")
    obj.text = "  "
    context = {"rst_mode": True}
    result = []
    _add_object_text_span(obj, context, result)
    assert result == ["  "]


def test_add_object_text_span_no_style_name():
    obj = Element.from_tag("text:span")
    obj.text = "text"
    context = {"rst_mode": True}
    result = []
    _add_object_text_span(obj, context, result)
    assert result == ["text"]


def test_add_object_text_span_empty_style_name_2():
    obj = Element.from_tag("text:span")
    obj.text = "text"
    obj.set_attribute("text:style-name", "")
    context = {"rst_mode": True}
    result = []
    _add_object_text_span(obj, context, result)
    assert result == ["text"]


def test_add_object_text_note_foot_no_citation():
    note = Note("footnote", citation=None, body="body")
    context = {"footnotes": [], "rst_mode": False}
    result = []
    _add_object_text(note, context, result)
    assert context["footnotes"] == [(0, "body")]
    assert result == ["[0]"]


def test_add_object_text_note_foot_rst():
    note = Note("footnote", citation="1", body="body")
    context = {"footnotes": [], "rst_mode": True}
    result = []
    _add_object_text_note_foot(note, context, result)
    assert result == [" [#]_ "]


def test_add_object_text_note_end_no_citation():
    note = Note("endnote", citation=None, body="body")
    context = {"endnotes": [], "rst_mode": False}
    result = []
    _add_object_text_note_end(note, context, result)
    assert context["endnotes"] == [(0, "body")]
    assert result == ["(0)"]


def test_add_object_text_note_end_rst():
    note = Note("endnote", citation="1", body="body")
    context = {"endnotes": [], "rst_mode": True}
    result = []
    _add_object_text_note_end(note, context, result)
    assert result == [" [*]_ "]


def test_add_object_text_span_rst_bold():
    obj = Element.from_tag("text:span")
    obj.text = "bold"
    obj.set_attribute("text:style-name", "bold_style")

    mock_doc = MagicMock()
    mock_style = MagicMock()
    mock_style.get_properties.return_value = {"fo:font-weight": "bold"}
    mock_doc.get_style.return_value = mock_style

    context = {"rst_mode": True, "document": mock_doc}
    result = []
    _add_object_text_span(obj, context, result)
    assert result == ["**bold**"]


def test_add_object_text_span_rst_italic():
    obj = Element.from_tag("text:span")
    obj.text = "italic"
    obj.set_attribute("text:style-name", "italic_style")

    mock_doc = MagicMock()
    mock_style = MagicMock()
    mock_style.get_properties.return_value = {"fo:font-style": "italic"}
    mock_doc.get_style.return_value = mock_style

    context = {"rst_mode": True, "document": mock_doc}
    result = []
    _add_object_text_span(obj, context, result)
    assert result == ["*italic*"]


def test_add_object_text_span_rst_no_props():
    obj = Element.from_tag("text:span")
    obj.text = "normal"
    obj.set_attribute("text:style-name", "normal_style")

    mock_doc = MagicMock()
    mock_style = MagicMock()
    mock_style.get_properties.return_value = {}
    mock_doc.get_style.return_value = mock_style

    context = {"rst_mode": True, "document": mock_doc}
    result = []
    _add_object_text_span(obj, context, result)
    assert result == ["normal"]


def test_add_object_text_span_rst_no_doc():
    obj = Element.from_tag("text:span")
    obj.text = "normal"
    obj.set_attribute("text:style-name", "normal_style")

    context = {"rst_mode": True, "document": None}
    result = []
    _add_object_text_span(obj, context, result)
    assert result == ["normal"]


def test_add_object_text_annotation():
    obj = Annotation(body="note")
    obj.note_body = "note"
    context = {"annotations": [], "rst_mode": False}
    result = []
    _add_object_text(obj, context, result)
    assert context["annotations"] == ["note"]
    assert result == ["[*]"]


def test_add_object_text_annotation_rst():
    obj = Annotation(body="note")
    obj.note_body = "note"
    context = {"annotations": [], "rst_mode": True}
    result = []
    _add_object_text_annotation(obj, context, result)
    assert result == [" [#]_ "]


def test_add_object_text_tab():
    obj = Element.from_tag("text:tab")
    result = []
    _add_object_text(obj, {}, result)
    assert result == ["\t"]


def test_add_object_text_line_break():
    obj = Element.from_tag("text:line-break")
    result = []
    _add_object_text(obj, {"rst_mode": False}, result)
    assert result == ["\n"]


def test_add_object_text_line_break_rst():
    obj = Element.from_tag("text:line-break")
    result = []
    _add_object_text_line_break(obj, {"rst_mode": True}, result)
    assert result == ["\n|"]


def test_add_object_text_paragraph_tag():
    obj = Element.from_tag("text:p")
    obj.text = "text"
    result = []
    _add_object_text(obj, {}, result)
    assert result == ["text"]


def test_add_object_text_link_tag():
    obj = Element.from_tag("text:a")
    obj.text = "link"
    result = []
    _add_object_text(obj, {}, result)
    assert result == ["link"]


def test_add_object_text_other_tag():
    obj = Header(1, "title")
    result = []
    _add_object_text(obj, {}, result)
    assert result == ["title"]


def test_add_object_text_note_end_dispatch():
    note = Note("endnote", citation="1", body="body")
    context = {"endnotes": [], "rst_mode": False}
    result = []
    _add_object_text_note(note, context, result)
    assert result == ["(1)"]


def test_get_formatted_text_default_context():
    elem = DummyElement.from_tag("text:p")
    elem.text = "test"
    assert elem.get_formatted_text() == "test\n\n"
    assert elem.get_formatted_text(simple=True) == "test"


def test_formatted_text_etext():
    elem = Element.from_tag("text:p")
    elem.append("hello")
    res = _formatted_text(elem, {})
    assert res == "hello"


def test_get_formatted_text_with_context():
    elem = DummyElement.from_tag("text:p")
    elem.text = "test"
    context = {"rst_mode": True, "img_counter": 0, "images": []}
    assert elem.get_formatted_text(context=context, simple=True) == "test"


def test_add_object_text_span_hasattr_branches():
    context = {"rst_mode": True}
    result = []
    obj1 = Element.from_tag("text:span")
    obj1.set_attribute("text:style-name", "S1")
    _add_object_text_span(obj1, context, result)
    assert result == [""]  # text is empty

    class NoStyleElement:
        tag = "text:span"

        def xpath(self, query):
            return ["content"]

    obj2 = NoStyleElement()
    with patch(
        "odfdo.mixin_paragraph_formatted._formatted_text",
        return_value="text",
    ):
        _add_object_text_span(obj2, context, result)
    assert result[1] == "text"


def test_formatted_text_full_recursive():
    elem = Element.from_tag("text:p")
    span = Element.from_tag("text:span")
    span.text = "inner"
    elem.append(span)
    res = _formatted_text(elem, {})
    assert res == "inner"


def test_add_object_text_span_italic_else_branch():
    obj = Element.from_tag("text:span")
    obj.text = "text"
    obj.set_attribute("text:style-name", "style")

    mock_doc = MagicMock()
    mock_style = MagicMock()
    # Properties exists, but neither bold nor italic
    mock_style.get_properties.return_value = {"fo:color": "red"}
    mock_doc.get_style.return_value = mock_style

    context = {"rst_mode": True, "document": mock_doc}
    result = []
    _add_object_text_span(obj, context, result)
    assert result == ["text"]



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
from __future__ import annotations

from collections import namedtuple
from collections.abc import Iterable
from datetime import datetime

import pytest

from odfdo.document import Document
from odfdo.header import Header
from odfdo.paragraph import Paragraph
from odfdo.tracked_changes import (
    ChangeInfo,
    TextChange,
    TextChangedRegion,
    TextChangeEnd,
    TextChangeStart,
    TextDeletion,
    TextFormatChange,
    TextInsertion,
    TrackedChanges,
)

Sample = namedtuple("Sample", ["doc", "changes"])


@pytest.fixture
def sample(samples) -> Iterable[Sample]:
    document = Document(samples("tracked_changes.odt"))
    yield Sample(
        doc=document,
        changes=document.body.get_tracked_changes(),
    )


@pytest.fixture
def sample_doc() -> Iterable[Document]:
    doc = Document("odt")
    body = doc.body
    body.clear()
    body.append(Header(1, "Title"))
    body.append(Paragraph("Some text."))
    yield doc


def test_change_info_class():
    ci = ChangeInfo()
    assert isinstance(ci, ChangeInfo)


def test_change_info_class_creator():
    name = "John Doe"
    dt = datetime(1977, 12, 25, 12, 0, 0)
    ci = ChangeInfo(name, dt)
    assert ci.creator == name


def test_change_info_class_date():
    name = "John Doe"
    dt = datetime(1977, 12, 25, 12, 0, 0)
    ci = ChangeInfo(name, dt)
    assert ci.date == dt


def test_change_info_comments_none():
    name = "John Doe"
    dt = datetime(1977, 12, 25, 12, 0, 0)
    ci = ChangeInfo(name, dt)
    assert ci.get_comments() == ""


def test_change_info_comments():
    name = "John Doe"
    dt = datetime(1977, 12, 25, 12, 0, 0)
    comment = "Some comment"
    ci = ChangeInfo(name, dt)
    ci.set_comments(comment)
    assert ci.get_comments() == comment


def test_change_info_comments_replace():
    name = "John Doe"
    dt = datetime(1977, 12, 25, 12, 0, 0)
    comment1 = "Some comment"
    comment2 = "More comment"
    ci = ChangeInfo(name, dt)
    ci.set_comments(comment1)
    ci.set_comments(comment2)
    assert ci.get_comments() == comment2


def test_change_info_comments_no_replace():
    name = "John Doe"
    dt = datetime(1977, 12, 25, 12, 0, 0)
    comment1 = "Some comment"
    comment2 = "More comment"
    ci = ChangeInfo(name, dt)
    ci.set_comments(comment1)
    ci.set_comments(comment2, replace=False)
    expected = f"{comment1}\n{comment2}"
    assert ci.get_comments() == expected


def test_change_info_comments_no_replace_no_join():
    name = "John Doe"
    dt = datetime(1977, 12, 25, 12, 0, 0)
    comment1 = "Some comment"
    comment2 = "More comment"
    ci = ChangeInfo(name, dt)
    ci.set_comments(comment1)
    ci.set_comments(comment2, replace=False)
    expected = [comment1, comment2]
    assert ci.get_comments(joined=False) == expected


def test_text_insertion_class():
    ti = TextInsertion()
    assert isinstance(ti, TextInsertion)


def test_text_insertion_empty_deleted():
    ti = TextInsertion()
    assert ti.get_deleted() is None


def test_text_insertion_empty_deleted_as_text():
    ti = TextInsertion()
    assert ti.get_deleted(as_text=True) == ""


def test_text_insertion_empty_inserted():
    ti = TextInsertion()
    with pytest.raises(TypeError):
        ti.get_inserted()


def test_text_insertion_empty_change_info():
    ti = TextInsertion()
    assert ti.get_change_info() is None


def test_text_insertion_empty_set_change_info():
    ti = TextInsertion()
    name = "John Doe"
    dt = datetime(1977, 12, 25, 12, 0, 0)
    comment = Paragraph("Some comment")
    ti.set_change_info(None, name, dt, comment)
    ci = ti.get_change_info()
    assert isinstance(ci, ChangeInfo)


def test_text_insertion_empty_set_change_info_comment_list():
    ti = TextInsertion()
    name = "John Doe"
    dt = datetime(1977, 12, 25, 12, 0, 0)
    comments = [Paragraph("Some comment"), Paragraph("Other comment")]
    ti.set_change_info(None, name, dt, comments)
    ci = ti.get_change_info()
    assert isinstance(ci, ChangeInfo)


def test_text_insertion_empty_set_change_info_get():
    ti = TextInsertion()
    name = "John Doe"
    dt = datetime(1977, 12, 25, 12, 0, 0)
    comment = Paragraph("Some comment")
    ti.set_change_info(None, name, dt, comment)
    ci = ti.get_change_info()
    assert ci.creator == name
    assert ci.date == dt
    assert ci.get_comments() == str(comment).strip()


def test_text_insertion_empty_set_change_info_get_no_comment():
    ti = TextInsertion()
    name = "John Doe"
    dt = datetime(1977, 12, 25, 12, 0, 0)
    ti.set_change_info(None, name, dt)
    ci = ti.get_change_info()
    assert ci.creator == name
    assert ci.date == dt
    assert ci.get_comments() == ""


def test_text_insertion_empty_set_change_info_get_comment_list():
    ti = TextInsertion()
    name = "John Doe"
    dt = datetime(1977, 12, 25, 12, 0, 0)
    t1 = "Some comment"
    t2 = "Other comment"
    comments = [Paragraph(t1), Paragraph(t2)]
    ti.set_change_info(None, name, dt, comments)
    ci = ti.get_change_info()
    assert ci.creator == name
    assert ci.date == dt
    assert ci.get_comments(joined=False) == [t1, t2]


def test_text_insertion_empty_set_change_info_get_comment_list_wrong():
    ti = TextInsertion()
    name = "John Doe"
    dt = datetime(1977, 12, 25, 12, 0, 0)
    t1 = "Some comment"
    comments = [Paragraph(t1), "wrong"]
    with pytest.raises(TypeError):
        ti.set_change_info(None, name, dt, comments)


def test_text_insertion_get_from_text_changed_region(sample):
    tracked_changes = sample.changes
    cr = tracked_changes.get_changed_region(1)
    assert isinstance(cr, TextChangedRegion)
    ti = cr.get_change_element()
    assert isinstance(ti, TextInsertion)


def test_text_insertion_get_inserted(sample):
    tracked_changes = sample.changes
    cr = tracked_changes.get_changed_region(1)
    ti = cr.get_change_element()
    content = ti.get_inserted()
    assert isinstance(content, list)
    p = content[0]
    assert isinstance(p, Paragraph)
    assert str(p).strip() == "los bonitos amigos"


def test_text_insertion_get_inserted_as_text(sample):
    tracked_changes = sample.changes
    cr = tracked_changes.get_changed_region(1)
    ti = cr.get_change_element()
    content = ti.get_inserted(as_text=True)
    assert content.strip() == "los bonitos amigos"


def test_text_insertion_get_inserted_as_text_no_header(sample):
    tracked_changes = sample.changes
    cr = tracked_changes.get_changed_region(1)
    ti = cr.get_change_element()
    content = ti.get_inserted(as_text=True, no_header=True)
    assert content.strip() == "los bonitos amigos"


def test_text_insertion_set_ci_wrong(sample):
    tracked_changes = sample.changes
    cr = tracked_changes.get_changed_region(1)
    ti = cr.get_change_element()
    with pytest.raises(TypeError):
        ti.set_change_info(change_info="wrong")


def test_text_insertion_set_ci_new(sample):
    tracked_changes = sample.changes
    cr = tracked_changes.get_changed_region(1)
    ti = cr.get_change_element()
    name = "John Doe"
    dt = datetime(1977, 12, 25, 12, 0, 0)
    ci = ChangeInfo(name, dt)
    ti.set_change_info(change_info=ci)
    new_ci = ti.get_change_info()
    assert new_ci.creator == name
    assert new_ci.date == dt


def test_text_deletion_class():
    td = TextDeletion()
    assert isinstance(td, TextDeletion)


def test_text_deletion_empty_deleted():
    td = TextDeletion()
    assert td.get_deleted() == []


def test_text_deletion_empty_deleted_as_text():
    td = TextDeletion()
    assert td.get_deleted(as_text=True) == ""


def test_text_deletion_empty_deleted_no_header():
    td = TextDeletion()
    assert td.get_deleted(no_header=True) == []


def test_text_deletion_empty_inserted():
    td = TextDeletion()
    assert td.get_inserted() is None


def test_text_deletion_empty_inserted_as_text():
    td = TextDeletion()
    assert td.get_inserted(as_text=True) == ""


def test_text_deletion_empty_change_info():
    td = TextDeletion()
    assert td.get_change_info() is None


def test_text_deletion_get_from_text_changed_region(sample):
    tracked_changes = sample.changes
    cr = tracked_changes.get_changed_region(0)
    assert isinstance(cr, TextChangedRegion)
    td = cr.get_change_element()
    assert isinstance(td, TextDeletion)


def test_text_deletion_get_deleted_from_text_changed_region(sample):
    tracked_changes = sample.changes
    cr = tracked_changes.get_changed_region(0)
    td = cr.get_change_element()
    assert len(td.get_deleted()) == 1
    assert str(td.get_deleted()[0]).strip() == "les"


def test_text_deletion_get_deleted_from_text_changed_region_as_text(sample):
    tracked_changes = sample.changes
    cr = tracked_changes.get_changed_region(0)
    td = cr.get_change_element()
    assert td.get_deleted(as_text=True).strip() == "les"


def test_text_deletion_get_deleted_from_text_changed_region_no_header(sample):
    tracked_changes = sample.changes
    cr = tracked_changes.get_changed_region(0)
    td = cr.get_change_element()
    assert str(td.get_deleted(no_header=True)[0]).strip() == "les"


def test_text_deletion_set_deleted(sample_doc):
    body = sample_doc.body
    p = body.get_paragraph(content="Some")
    assert isinstance(p, Paragraph)
    td = TextDeletion()
    assert len(td.get_deleted()) == 0
    td.set_deleted(p)
    assert len(td.get_deleted()) == 1


def test_text_deletion_set_deleted_list(sample_doc):
    body = sample_doc.body
    p = body.get_paragraph(content="Some")
    assert isinstance(p, Paragraph)
    td = TextDeletion()
    assert len(td.get_deleted()) == 0
    td.set_deleted([p])
    assert len(td.get_deleted()) == 1


def test_text_deletion_set_deleted_list_2(sample_doc):
    body = sample_doc.body
    h = body.get_header(content="Title")
    assert isinstance(h, Header)
    td = TextDeletion()
    assert len(td.get_deleted()) == 0
    td.set_deleted([h])
    assert len(td.get_deleted()) == 1


def test_text_deletion_set_deleted_list_3(sample_doc):
    body = sample_doc.body
    h = body.get_header(content="Title")
    assert isinstance(h, Header)
    p = body.get_paragraph(content="Some")
    assert isinstance(p, Paragraph)
    td = TextDeletion()
    assert len(td.get_deleted()) == 0
    td.set_deleted([h, p])
    assert len(td.get_deleted()) == 2


def test_text_deletion_set_deleted_list_2_replace(sample_doc):
    body = sample_doc.body
    h = body.get_header(content="Title")
    assert isinstance(h, Header)
    p = body.get_paragraph(content="Some")
    assert isinstance(p, Paragraph)
    td = TextDeletion()
    assert len(td.get_deleted()) == 0
    td.set_deleted([h])
    assert len(td.get_deleted()) == 1
    td.set_deleted([p])
    assert len(td.get_deleted()) == 1


def test_text_deletion_set_deleted_get_no_header(sample_doc):
    body = sample_doc.body
    h = body.get_header(content="Title")
    assert isinstance(h, Header)
    p = body.get_paragraph(content="Some")
    assert isinstance(p, Paragraph)
    td = TextDeletion()
    assert len(td.get_deleted()) == 0
    td.set_deleted([h, p])
    result = td.get_deleted(no_header=True)
    assert len(result) == 2
    assert isinstance(result[0], Paragraph)
    assert isinstance(result[1], Paragraph)


def test_text_format_change_class():
    tfc = TextFormatChange()
    assert isinstance(tfc, TextFormatChange)


def test_text_changed_region_class():
    tcr = TextChangedRegion()
    assert isinstance(tcr, TextChangedRegion)


def test_text_changed_region_get_changed_region(sample):
    trc = sample.doc.body.get_tracked_changes()
    assert isinstance(trc, TrackedChanges)
    tcr = trc.get_changed_region(0)
    assert isinstance(tcr, TextChangedRegion)


def test_text_changed_region_get_change_info(sample):
    trc = sample.doc.body.get_tracked_changes()
    tcr = trc.get_changed_region(0)
    ci = tcr.get_change_info()
    assert isinstance(ci, ChangeInfo)


def test_text_changed_region_set_change_info(sample):
    trc = sample.doc.body.get_tracked_changes()
    tcr = trc.get_changed_region(0)
    name = "John Doe"
    dt = datetime(1977, 12, 25, 12, 0, 0)
    new_ci = ChangeInfo(name, dt)
    tcr.set_change_info(new_ci)
    result_ci = tcr.get_change_info()
    assert result_ci.creator == name


def test_text_changed_region_set_change_info_wrong(sample):
    tcr = TextChangedRegion()
    name = "John Doe"
    dt = datetime(1977, 12, 25, 12, 0, 0)
    new_ci = ChangeInfo(name, dt)
    with pytest.raises(ValueError):
        tcr.set_change_info(new_ci)


def test_text_changed_region_set_id(sample):
    trc = sample.doc.body.get_tracked_changes()
    tcr = trc.get_changed_region(0)
    tcr.set_id("123")
    assert tcr.get_id() == "123"


def test_text_changed_region_get_xml_id(sample):
    trc = sample.doc.body.get_tracked_changes()
    tcr = trc.get_changed_region(0)
    assert tcr._get_xml_id() == "ct140266191788864"


def test_tracked_changes_class():
    tcs = TrackedChanges()
    assert isinstance(tcs, TrackedChanges)


def test_tracked_changes_per_role_deletion(sample):
    trc = sample.doc.body.get_tracked_changes()
    regions = trc.get_changed_regions(role="deletion")
    assert len(regions) == 2


def test_tracked_changes_per_role_insertion(sample):
    trc = sample.doc.body.get_tracked_changes()
    regions = trc.get_changed_regions(role="insertion")
    assert len(regions) == 1


def test_tracked_changes_per_role_insertion_empty(sample):
    trc = TrackedChanges()
    regions = trc.get_changed_regions(role="insertion")
    assert len(regions) == 0


def test_text_change_class():
    tc = TextChange()
    assert isinstance(tc, TextChange)


def test_text_change_class_base_class_1():
    tc = TextChange()
    assert tc.get_inserted() is None


def test_text_change_class_base_class_2():
    tc = TextChange()
    assert tc.get_start() is None


def test_text_change_class_base_class_3():
    tc = TextChange()
    assert tc.get_end() is None


def test_text_change_class_base_class_4():
    tc = TextChange()
    with pytest.raises(ValueError):
        tc.get_changed_region()


def test_text_change_get(sample):
    body = sample.doc.body
    changes = body.get_text_changes()
    assert len(changes) == 3


def test_text_changes_property(sample):
    body = sample.doc.body
    changes = body.text_changes
    assert len(changes) == 3


def test_get_text_change(sample):
    body = sample.doc.body
    change = body.get_text_change(0)
    assert isinstance(change, TextChange)
    assert change.get_id() == "ct140266191788864"


def test_get_text_change_idx(sample):
    body = sample.doc.body
    change = body.get_text_change(idx="ct140266191788864")
    assert isinstance(change, TextChange)
    assert change.get_id() == "ct140266191788864"


def test_get_text_change_idx_start(sample):
    body = sample.doc.body
    # ct140266193096576 is a change-start in the sample
    change = body.get_text_change(idx="ct140266193096576")
    assert isinstance(change, TextChangeStart)
    assert change.get_id() == "ct140266193096576"


def test_get_text_change_none(sample_doc):
    body = sample_doc.body
    change = body.get_text_change(0)
    assert change is None


def test_get_text_change_idx_none(sample_doc):
    body = sample_doc.body
    change = body.get_text_change(idx="nonexistent")
    assert change is None


def test_text_change_get_id(sample):
    body = sample.doc.body
    change = body.get_text_changes()[0]
    assert change.get_id() == "ct140266191788864"


def test_text_change_set_id(sample):
    body = sample.doc.body
    change = body.get_text_changes()[0]
    change.set_id("aaa123")
    assert change.get_id() == "aaa123"


def test_text_get_changed_region(sample):
    body = sample.doc.body
    change = body.get_text_changes()[0]
    changed_region = change.get_changed_region()
    assert isinstance(changed_region, TextChangedRegion)


def test_text_get_changed_region_2(sample):
    body = sample.doc.body
    change = body.get_text_changes()[0]
    tracked_changes = body.get_tracked_changes()
    changed_region = change.get_changed_region(tracked_changes)
    assert isinstance(changed_region, TextChangedRegion)


def test_text_get_change_info(sample):
    body = sample.doc.body
    change = body.get_text_changes()[0]
    ci = change.get_change_info()
    assert isinstance(ci, ChangeInfo)
    assert ci.creator == "Romain Gauthier"


def test_text_get_change_info_2(sample):
    body = sample.doc.body
    change = body.get_text_changes()[0]
    tracked_changes = body.get_tracked_changes()
    ci = change.get_change_info(tracked_changes)
    assert isinstance(ci, ChangeInfo)
    assert ci.creator == "Romain Gauthier"


def test_text_get_change_element_1(sample):
    body = sample.doc.body
    change = body.get_text_changes()[0]
    # tracked_changes = body.get_tracked_changes()
    ce = change.get_change_element()
    assert isinstance(ce, TextDeletion)


def test_text_get_change_element_2(sample):
    body = sample.doc.body
    change = body.get_text_changes()[0]
    tracked_changes = body.get_tracked_changes()
    ce = change.get_change_element(tracked_changes)
    assert isinstance(ce, TextDeletion)


def test_text_get_change_element_3(sample):
    body = sample.doc.body
    change = body.get_text_changes()[1]
    ce = change.get_change_element()
    assert isinstance(ce, TextInsertion)


def test_text_get_change_element_4(sample):
    body = sample.doc.body
    change = body.get_text_changes()[2]
    ce = change.get_change_element()
    assert isinstance(ce, TextDeletion)


def test_text_get_deleted_1(sample):
    body = sample.doc.body
    change = body.get_text_changes()[0]
    ce = change.get_change_element()
    deleted = ce.get_deleted()
    assert str(deleted[0]).strip() == "les"


def test_text_get_deleted_2(sample):
    body = sample.doc.body
    change = body.get_text_changes()[1]
    ce = change.get_change_element()
    deleted = ce.get_deleted()
    assert deleted is None


def test_text_get_deleted_3(sample):
    body = sample.doc.body
    change = body.get_text_changes()[2]
    ce = change.get_change_element()
    deleted = ce.get_deleted()
    assert str(deleted[0]).strip() == "amis,"


def test_text_change_end_class():
    tce = TextChangeEnd()
    assert isinstance(tce, TextChangeEnd)


def test_text_change_end_none_start():
    tce = TextChangeEnd()
    assert tce.get_start() is None


def test_text_change_end_get_end():
    tce = TextChangeEnd()
    assert tce.get_end() == tce


def test_text_change_end_get_deleted():
    tce = TextChangeEnd()
    assert tce.get_deleted() is None


def test_text_change_end_get_inserted_0():
    tce = TextChangeEnd()
    assert tce.get_inserted() is None


def test_text_change_end_get_inserted_1():
    tce = TextChangeEnd()
    assert tce.get_inserted(as_text=True) == ""


def test_text_change_start_class():
    tcst = TextChangeStart()
    assert isinstance(tcst, TextChangeStart)


def test_text_change_start_none_start():
    tcst = TextChangeStart()
    assert tcst.get_start() == tcst


def test_text_change_start_get_end():
    tcst = TextChangeStart()
    assert tcst.get_end() is None


def test_text_change_start_elem_1(sample):
    body = sample.doc.body
    starts = body.get_text_change_starts()
    assert len(starts) == 1
    assert isinstance(starts[0], TextChangeStart)


def test_text_change_start_elem_2(sample):
    body = sample.doc.body
    start = body.get_text_change_start(0)
    assert isinstance(start, TextChangeStart)


def test_text_change_start_elem_3(sample):
    body = sample.doc.body
    start = body.get_text_change_start(42)
    assert start is None


def test_text_change_end_elem_1(sample):
    body = sample.doc.body
    ends = body.get_text_change_ends()
    assert len(ends) == 1
    assert isinstance(ends[0], TextChangeEnd)


def test_text_change_end_elem_2(sample):
    body = sample.doc.body
    end = body.get_text_change_end(0)
    assert isinstance(end, TextChangeEnd)


def test_text_change_end_elem_3(sample):
    body = sample.doc.body
    end = body.get_text_change_end(42)
    assert end is None


def test_text_change_start_delete(sample):
    body = sample.doc.body
    start = body.get_text_change_start(0)
    start.delete()
    start2 = body.get_text_change_start(0)
    assert start2 is None
    end2 = body.get_text_change_end(0)
    assert end2 is None


def test_text_change_start_get_deleted(sample):
    body = sample.doc.body
    start = body.get_text_change(0)
    assert isinstance(start.get_deleted()[0], Paragraph)


def test_get_text_change_deletions(sample):
    body = sample.doc.body
    deletions = body.get_text_change_deletions()
    assert len(deletions) == 2
    assert isinstance(deletions[0], TextChange)


def test_get_text_change_deletion(sample):
    body = sample.doc.body
    deletion = body.get_text_change_deletion(0)
    assert isinstance(deletion, TextChange)
    assert deletion.get_id() == "ct140266191788864"


def test_get_text_change_deletion_idx(sample):
    body = sample.doc.body
    deletion = body.get_text_change_deletion(idx="ct140266191788864")
    assert isinstance(deletion, TextChange)
    assert deletion.get_id() == "ct140266191788864"


def test_get_text_change_deletions_empty(sample_doc):
    body = sample_doc.body
    deletions = body.get_text_change_deletions()
    assert deletions == []


def test_get_text_change_deletion_none(sample_doc):
    body = sample_doc.body
    deletion = body.get_text_change_deletion(0)
    assert deletion is None


def test_get_text_change_deletion_idx_none(sample_doc):
    body = sample_doc.body
    deletion = body.get_text_change_deletion(idx="nonexistent")
    assert deletion is None
